"""A numpy version of Powderworld simulator.

The code is based on the original Powderworld simulator written in PyTorch. It is supposed to work identically to the
original simulator, but for performance reasons, the update rules that are not used in the OGBench Powderworld tasks are
commented out. To use the full simulator, uncomment the unused update rules in the `register_update_rules` function.
"""

from collections import namedtuple

import numpy as np

Info = namedtuple('Info', ['rand_movement', 'rand_interact', 'rand_element'])

# ================ REGISTER ELEMENTS. =================
# Name:    ID, Density, GravityInter
pw_elements = {
    'empty': (0, 1, 1),
    'wall': (1, 4, 0),
    'sand': (2, 3, 1),
    'water': (3, 2, 1),
    'gas': (4, 0, 1),
    'wood': (5, 4, 0),
    'ice': (6, 4, 0),
    'fire': (7, 0, 1),
    'plant': (8, 4, 0),
    'stone': (9, 3, 1),
    'lava': (10, 3, 1),
    'acid': (11, 2, 1),
    'dust': (12, 2, 1),
    'cloner': (13, 4, 0),
    'agentFish': (14, 2, 1),
    'agentBird': (15, 4, 0),
    'agentKangaroo': (16, 3, 1),
    'agentMole': (17, 3, 1),
    'agentLemming': (18, 3, 1),
    'agentSnake': (19, 4, 0),
    'agentRobot': (20, 3, 1),
}

pw_element_names = [
    'empty',
    'wall',
    'sand',
    'water',
    'gas',
    'wood',
    'ice',
    'fire',
    'plant',
    'stone',
    'lava',
    'acid',
    'dust',
    'cloner',
    'agentFish',
    'agentBird',
    'agentKangaroo',
    'agentMole',
    'agentLemming',
    'agentSnake',
    'agentRobot',
]


# ================================================
# ============= HELPERS ==================
# ================================================


def get_below(x):
    return np.roll(x, shift=-1, axis=2)


def get_above(x):
    return np.roll(x, shift=1, axis=2)


def get_left(x):
    return np.roll(x, shift=1, axis=3)


def get_right(x):
    return np.roll(x, shift=-1, axis=3)


def get_in_cardinal_direction(x, directions):
    y = get_right(x) * (directions == 0)
    y = y + get_below(x) * (directions == 2)
    y = y + get_left(x) * (directions == 4)
    y = y + get_above(x) * (directions == 6)
    return y


def interp(switch, if_false, if_true):
    return (~switch) * if_false + (switch) * if_true


def interp_int(switch, if_false, if_true: int):
    return (~switch) * if_false + (switch) * if_true


def interp2(switch_a, switch_b, if_false, if_a, if_b):
    return ((~switch_a) & (~switch_b)) * if_false + (switch_a) * if_a + (switch_b) * if_b


def interp_swaps8(swaps, world, w0, w1, w2, w3, w4, w5, w6, w7):
    new_world = world * (swaps == -1)
    new_world += w0 * (swaps == 0)
    new_world += w1 * (swaps == 1)
    new_world += w2 * (swaps == 2)
    new_world += w3 * (swaps == 3)
    new_world += w4 * (swaps == 4)
    new_world += w5 * (swaps == 5)
    new_world += w6 * (swaps == 6)
    new_world += w7 * (swaps == 7)
    return new_world


def interp_swaps4(swaps, world, w0, w1, w2, w3):
    new_world = world * (swaps == -1)
    new_world += w0 * (swaps == 0)
    new_world += w1 * (swaps == 1)
    new_world += w2 * (swaps == 2)
    new_world += w3 * (swaps == 3)
    return new_world


def normalize(x, p=2, axis=0, eps=1e-12):
    norm = np.linalg.norm(x, ord=p, axis=axis, keepdims=True)
    return x / (norm + eps)


# Source: https://github.com/99991/NumPyConv2D
def conv2d(input, weight, bias=None, stride=1, padding=0, dilation=1, groups=1, padding_mode='zeros'):
    """
    Applies a 2D convolution to an array of images. Technically, this function
    computes a correlation instead of a convolution since the kernels are not
    flipped.

    input: numpy array of images with shape = (n, c, h, w)
    weight: numpy array with shape = (c_out, c // groups, kernel_height, kernel_width)
    bias: numpy array of shape (c_out,), default None
    stride: step width of convolution kernel, int or (int, int) tuple, default 1
    padding: padding around images, int or (int, int) tuple or "same", default 0
    dilation: spacing between kernel elements, int or (int, int) tuple, default 1
    groups: split c and c_out into groups to reduce memory usage (must both be divisible), default 1
    padding_mode: "zeros", "reflect", "replicate", "circular", or whatever np.pad supports, default "zeros"

    This function is indended to be similar to PyTorch's conv2d function.
    For more details, see:
    https://pytorch.org/docs/stable/generated/torch.nn.functional.conv2d.html
    https://pytorch.org/docs/stable/generated/torch.nn.Conv2d.html#torch.nn.Conv2d
    """
    c_out, c_in_by_groups, kh, kw = weight.shape
    kernel_size = (kh, kw)

    if isinstance(stride, int):
        stride = [stride, stride]

    if isinstance(dilation, int):
        dilation = [dilation, dilation]

    if padding:
        input = conv2d_pad(input, padding, padding_mode, stride, dilation, kernel_size)

    n, c_in, h, w = input.shape
    dh, dw = dilation
    sh, sw = stride
    dilated_kh = (kh - 1) * dh + 1
    dilated_kw = (kw - 1) * dw + 1
    assert c_in % groups == 0, f'Number of input channels ({c_in}) not divisible by groups ({groups}).'
    assert c_out % groups == 0, f'Number of output channels ({c_out}) not divisible by groups ({groups}).'
    c_in_group = c_in // groups
    c_out_group = c_out // groups
    kernel_shape = (c_in_group, dilated_kh, dilated_kw)

    input = input.reshape(n, groups, c_in_group, h, w)
    weight = weight.reshape(groups, c_out_group, c_in_by_groups, kh, kw)

    # Cut out kernel-shaped regions from input
    windows = np.lib.stride_tricks.sliding_window_view(input, kernel_shape, axis=(-3, -2, -1))

    # Apply stride and dilation. Prepare for broadcasting to handle groups.
    windows = windows[:, :, :, ::sh, ::sw, :, ::dh, ::dw]
    weight = weight[np.newaxis, :, :, np.newaxis, np.newaxis, :, :, :]
    h_out, w_out = windows.shape[3:5]

    # Dot product equivalent to either of the next two lines but 10 times faster
    # y = np.einsum("abcdeijk,abcdeijk->abcde", windows, weight)
    # y = (windows * weight).sum(axis=(-3, -2, -1))
    windows = windows.reshape(n, groups, 1, h_out, w_out, c_in_group * kh * kw)
    weight = weight.reshape(1, groups, c_out_group, 1, 1, c_in_group * kh * kw)
    y = np.einsum('abcdei,abcdei->abcde', windows, weight)

    # Concatenate groups as output channels
    y = y.reshape(n, c_out, h_out, w_out)

    if bias is not None:
        y = y + bias.reshape(1, c_out, 1, 1)

    return y


def conv2d_pad(input, padding, padding_mode, stride, dilation, kernel_size):
    if padding == 'valid':
        return input

    if padding == 'same':
        h, w = input.shape[-2:]
        sh, sw = stride
        dh, dw = dilation
        kh, kw = kernel_size
        ph = (h - 1) * (sh - 1) + (kh - 1) * dh
        pw = (w - 1) * (sw - 1) + (kw - 1) * dw
        ph0 = ph // 2
        ph1 = ph - ph0
        pw0 = pw // 2
        pw1 = pw - pw0
    else:
        if isinstance(padding, int):
            padding = [padding, padding]
        ph0, pw0 = padding
        ph1, pw1 = padding

    pad_width = ((0, 0), (0, 0), (ph0, ph1), (pw0, pw1))

    mode = {
        'zeros': 'constant',
        'reflect': 'reflect',
        'replicate': 'edge',
        'circular': 'wrap',
    }.get(padding_mode, padding_mode)

    return np.pad(input, pad_width, mode)


# ================================================
# ============= GENERAL CLASS =================
# ================================================


class PWSim:
    def __init__(self):
        self.elements = pw_elements
        self.element_names = pw_element_names

        self.NUM_ELEMENTS = len(self.elements)
        # [ElementID(0), Density(1), GravityInter(2), VelocityField(3, 4), Color(5), Custom1(6), Custom2(7), Custom3(8)]
        # Custom 1              Custom 2          Custom 3
        # BirdVelX              BirdVelY
        #                                         DidGravity
        # FluidMomentum
        # FluidMomentum         KangarooJump
        # MoleDirection
        # SnakeDirection.       SnakeEnergy
        self.NUM_CHANNEL = 1 + 1 + 1 + 2 + 1 + 3
        self.pw_type = np.float32

        # ================ NUMPY KERNELS =================
        self.elem_vecs = {}
        self.elem_vecs_array = np.zeros((self.NUM_ELEMENTS, self.NUM_CHANNEL), dtype=self.pw_type)
        for elem_name, elem in self.elements.items():
            elem_vec = np.zeros(self.NUM_CHANNEL)
            elem_vec[0] = elem[0]
            elem_vec[1] = elem[1]
            elem_vec[2] = elem[2]
            self.elem_vecs[elem_name] = elem_vec[None, :, None, None]
            self.elem_vecs_array[elem[0]] = elem_vec

        self.neighbor_kernel = np.ones((1, 1, 3, 3), dtype=self.pw_type)
        self.zero = np.zeros((1, 1), dtype=self.pw_type)
        self.one = np.ones((1, 1), dtype=self.pw_type)

        self.up = np.array([-1, 0])[None, :, None, None]
        self.down = np.array([1, 0])[None, :, None, None]
        self.left = np.array([0, -1])[None, :, None, None]
        self.right = np.array([0, 1])[None, :, None, None]

        self.register_update_rules()

    # ==================================================
    # ============ REGISTER UPDATE RULES ===============
    # ==================================================
    def register_update_rules(self):
        """
        Overwrite this function with your own set of update rules to change behavior.
        """
        self.update_rules = [
            BehaviorStone(self),
            # BehaviorMole(self),
            BehaviorGravity(self),
            BehaviorSand(self),
            # BehaviorLemming(self),
            BehaviorFluidFlow(self),
            BehaviorIce(self),
            BehaviorWater(self),
            BehaviorFire(self),
            BehaviorPlant(self),
            # BehaviorLava(self),
            # BehaviorAcid(self),
            # BehaviorCloner(self),
            # BehaviorFish(self),
            # BehaviorBird(self),
            # BehaviorKangaroo(self),
            # BehaviorSnake(self),
            BehaviorVelocity(self),
        ]  # Unused rules commented out for performance
        self.update_rules_jit = None

    # =========== WORLD EDITING HELPERS ====================
    def add_element(self, world_slice, element_name, wind=None):
        if isinstance(element_name, int):
            element_name = self.element_names[element_name]

        if element_name == 'wind':
            world_slice[:, 3:5] = wind
        else:
            world_slice[...] = self.elem_vecs[element_name]
            if element_name == 'agentSnake':
                world_slice[:, 7] = 1

    def add_element_rc(self, world_slice, rr, cc, element_name):
        if isinstance(element_name, int):
            element_name = self.element_names[element_name]
        world_slice[:, :, rr, cc] = self.elem_vecs[element_name]

    def id_to_pw(self, world_ids):
        world = self.elem_vecs_array[world_ids]
        world = np.transpose(world, (0, 3, 1, 2))
        return world

    def np_to_pw(self, np_world):
        np_world_ids = np_world.astype(int).copy()
        return self.id_to_pw(np_world_ids)

    # =========== UPDATE HELPERS ====================
    def get_elem(self, world, elemname):
        elem_id = self.elements[elemname][0]
        return (world[:, 0:1] == elem_id).astype(self.pw_type)

    def get_bool(self, world, elemname):
        elem_id = self.elements[elemname][0]
        return world[:, 0:1] == elem_id

    def direction_func(self, d, x):
        if d == 0:
            return get_right(x)
        elif d == 1:
            return get_right(get_below(x))
        elif d == 2:
            return get_below(x)
        elif d == 3:
            return get_left(get_below(x))
        elif d == 4:
            return get_left(x)
        elif d == 5:
            return get_left(get_above(x))
        elif d == 6:
            return get_above(x)
        elif d == 7:
            return get_right(get_above(x))

    def forward(self, world):
        # Helper Functions
        rand_movement = np.random.rand(world.shape[0], 1, world.shape[2], world.shape[3]).astype(
            self.pw_type
        )  # For gravity
        rand_interact = np.random.rand(world.shape[0], 1, world.shape[2], world.shape[3]).astype(
            self.pw_type
        )  # For gravity
        rand_element = np.random.rand(world.shape[0], 1, world.shape[2], world.shape[3]).astype(
            self.pw_type
        )  # For gravity

        info = (rand_movement, rand_interact, rand_element)

        for update_rule in self.update_rules:
            world = update_rule.forward(world, info)

        return world


# ================================================
# ============== RENDERER ========================
# ================================================
class PWRenderer:
    def __init__(self):
        pw_type = np.float32
        self.elem_vecs_array = np.zeros((len(pw_elements), 3), dtype=pw_type)
        self.elem_vecs_array = (
            np.array(
                [
                    [236, 240, 241],  # EMPTY #ECF0F1
                    [108, 122, 137],  # WALL #6C7A89
                    [243, 194, 58],  # SAND #F3C23A
                    [75, 119, 190],  # WATER #4B77BE
                    [179, 157, 219],  # GAS #875F9A
                    [202, 105, 36],  # WOOD #CA6924
                    [137, 196, 244],  # ICE #89C4F4
                    [249, 104, 14],  # FIRE #F9680E
                    [38, 194, 129],  # PLANT #26C281
                    [38, 67, 72],  # STONE #264348
                    [157, 41, 51],  # LAVA #9D2933
                    [176, 207, 120],  # ACID #B0CF78
                    [255, 179, 167],  # DUST #FFB3A7
                    [191, 85, 236],  # CLONER #BF55EC
                    [0, 229, 255],  # AGENT FISH #00E5FF
                    [61, 90, 254],  # AGENT BIRD #3D5AFE
                    [121, 85, 72],  # AGENT KANGAROO #795548
                    [56, 142, 60],  # AGENT MOLE #388E3C
                    [158, 157, 36],  # AGENT LEMMING #9E9D24
                    [198, 40, 40],  # AGENT SNAKE #C62828
                    [224, 64, 251],  # AGENT ROBOT #E040FB
                ],
                dtype=pw_type,
            )
            / 255.0
        )

        self.vector_color_kernel = np.array([200, 100, 100], dtype=np.float32)
        self.vector_color_kernel /= 255.0
        self.vector_color_kernel = self.vector_color_kernel[:, None, None]

    # ================ RENDERING ====================
    def forward(self, world):
        img = self.elem_vecs_array[world[0:1, 0].astype(int)][0].transpose(2, 0, 1)
        if world.shape[1] > 1:
            velocity_field = world[0, 3:5]
            velocity_field_magnitudes = np.linalg.norm(velocity_field, axis=0)[None]

            velocity_field_angles_raw = (1 / (2 * np.pi)) * np.arccos(
                velocity_field[1] / (velocity_field_magnitudes + 0.001)
            )
            is_y_lessthan_zero = velocity_field[0] < 0
            velocity_field_angles_raw = interp(
                switch=is_y_lessthan_zero,
                if_false=velocity_field_angles_raw,
                if_true=(1 - velocity_field_angles_raw),
            )
            velocity_field_angles = velocity_field_angles_raw

            velocity_field_colors = self.vector_color_kernel

            velocity_field_display = np.clip(velocity_field_magnitudes / 5, 0, 0.5)
            img = (1 - velocity_field_display) * img + velocity_field_display * velocity_field_colors
        img = np.clip(img, 0, 1)
        return img

    def render(self, world):
        img = self.forward(world)
        img = img.transpose(1, 2, 0)
        img = (img * 255).astype(np.uint8)
        return img


# =========================================================================
# ====================== CORE UPDATE BEHAVIORS ============================
# =========================================================================


class BehaviorGravity:
    """
    Run gravity procedure.
    Loop through each possible density (1-5).
    In kernel form, compute for each block:
        IF density == currDensity && density BELOW is less && both gravity-affected -> Become below.
        (Else)If density ABOVE is greater && density ABOVE == currDensity && both gravity-affected -> Become above.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return True

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info

        world[:, 8:9] = interp(switch=(world[:, 2:3] == 1), if_false=world[:, 8:9], if_true=self.pw.zero)

        above = get_above(world)
        below = get_below(world)

        is_density_below_less = below[:, 1:2] - world[:, 1:2] < 0
        is_gravity = world[:, 2:3] == 1
        is_gravity_below = below[:, 2:3] == 1

        is_center_and_below_gravity = is_gravity_below & is_gravity
        does_become_below = is_density_below_less & is_center_and_below_gravity
        does_become_above = get_above(does_become_below)

        has_overlap = does_become_below & does_become_above
        does_become_below_real = does_become_below & ~has_overlap
        does_become_above_real = get_above(does_become_below_real)

        world[:] = interp2(
            switch_a=does_become_below_real, switch_b=does_become_above_real, if_false=world, if_a=below, if_b=above
        )
        world[:, 8:9] = interp(switch=does_become_above_real, if_false=world[:, 8:9], if_true=self.pw.one)

        return world


class BehaviorSand:
    """
    Run sand-piling procedure.
    Loop over each piling block type. In kernel form, for each block:
        If dir=left and BELOW_LEFT density is less && both gravity-affected -> Become below-left.
        If ABOVE_RIGHT dir=left and ABOVE_RIGHT density is less && both gravity-affected -> Become above-right.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return True

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info
        fall_dir = rand_movement > 0.5
        not_did_gravity = ~(world[:, 8:9] > 0)
        for fallLeft in [True, False]:
            get_in_dir = get_left if fallLeft else get_right
            get_in_not_dir = get_right if fallLeft else get_left
            world_below_left = get_in_dir(get_below(world))
            world_above_right = get_in_not_dir(get_above(world))

            fall_dir = (rand_movement > 0.5) if fallLeft else (rand_movement <= 0.5)
            rand_above_right = get_in_not_dir(get_above(fall_dir))

            is_below_left_density_lower = (world[:, 1:2] - world_below_left[:, 1:2]) > 0
            is_above_right_density_higher = (world_above_right[:, 1:2] - world[:, 1:2]) > 0
            is_gravity = world[:, 2:3] == 1
            is_below_left_gravity = world_below_left[:, 2:3] == 1
            is_above_right_gravity = world_above_right[:, 2:3] == 1
            is_element = self.pw.get_bool(world, 'sand') | self.pw.get_bool(world, 'dust')
            is_above_right_element = self.pw.get_bool(world_above_right, 'sand') | self.pw.get_bool(
                world_above_right, 'dust'
            )

            is_matching_fall = fall_dir
            is_above_right_matching_fall = rand_above_right
            not_did_gravity = ~(world[:, 8:9] > 0)
            not_did_gravity_below_left = ~(world_below_left[:, 8:9] > 0)
            not_did_gravity_above_right = ~(world_above_right[:, 8:9] > 0)

            does_become_below_left = (
                is_element
                & not_did_gravity_below_left
                & is_matching_fall
                & is_below_left_density_lower
                & is_below_left_gravity
                & not_did_gravity
            )
            does_become_above_right = (
                is_above_right_element
                & not_did_gravity_above_right
                & is_above_right_matching_fall
                & is_above_right_density_higher
                & is_above_right_gravity
                & not_did_gravity
            )

            world[:] = interp2(
                switch_a=does_become_below_left,
                switch_b=does_become_above_right,
                if_false=world,
                if_a=world_below_left,
                if_b=world_above_right,
            )
        return world


class BehaviorStone:
    """Run stone-stability procedure. If a stone is next to two stones, turn gravity off. Otherwise, turn it on."""

    def __init__(self, pw):
        self.pw = pw
        self.stone_kernel = np.zeros((1, 1, 3, 3), dtype=self.pw.pw_type)
        self.stone_kernel[0, 0, 0, 0] = 1
        self.stone_kernel[0, 0, 0, 2] = 1

    def check_filter(self, world):
        return True

    def forward(self, world, info):
        stone = self.pw.get_elem(world, 'stone')
        has_stone_supports = conv2d(stone, self.stone_kernel, padding=1)
        world[:, 2:3] = (1 - stone) * world[:, 2:3] + stone * (has_stone_supports < 2)
        return world


class BehaviorFluidFlow:
    """
    Run fluid-flowing procedure. Same as sand-piling, but move LEFT/RIGHT instead of BELOW-LEFT/BELOW-RIGHT.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return True

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info
        new_fluid_momentum = np.zeros((world.shape[0], 1, world.shape[2], world.shape[3]), dtype=self.pw.pw_type)
        for fallLeft in [True, False]:
            get_in_dir = get_left if fallLeft else get_right
            get_in_not_dir = get_right if fallLeft else get_left
            fall_dir = (rand_movement + world[:, 6:7] + new_fluid_momentum) > 0.5
            is_matching_fall = fall_dir if fallLeft else (~fall_dir)
            world_left = get_in_dir(world)
            world_right = get_in_not_dir(world)

            is_air_move = self.pw.get_bool(world, 'agentKangaroo') | self.pw.get_bool(world, 'agentLemming')
            is_element = (
                self.pw.get_bool(world, 'empty')
                | self.pw.get_bool(world, 'water')
                | self.pw.get_bool(world, 'gas')
                | self.pw.get_bool(world, 'lava')
                | self.pw.get_bool(world, 'acid')
                | is_air_move
            )
            is_left_density_lower = (world[:, 1:2] - world_left[:, 1:2]) > 0
            is_gravity = world[:, 2:3] == 1
            is_left_gravity = world_left[:, 2:3] == 1
            not_did_gravity_left = ~(world[:, 8:9] > 0) | is_air_move

            does_become_left = (
                is_matching_fall
                & is_element
                & not_did_gravity_left
                & is_left_density_lower
                & is_left_gravity
                & is_gravity
            )
            does_become_right = get_in_not_dir(does_become_left)

            has_overlap = does_become_left & does_become_right
            does_become_left_real = does_become_left & ~has_overlap
            does_become_right_real = get_in_not_dir(does_become_left_real)

            new_fluid_momentum += does_become_right_real * (2 if fallLeft else -2)

            world[:] = interp2(
                switch_a=does_become_left_real,
                switch_b=does_become_right_real,
                if_false=world,
                if_a=world_left,
                if_b=world_right,
            )

        world[:, 6:7, :, :] = interp(
            switch=(
                self.pw.get_bool(world, 'empty')
                | self.pw.get_bool(world, 'water')
                | self.pw.get_bool(world, 'gas')
                | self.pw.get_bool(world, 'lava')
                | self.pw.get_bool(world, 'acid')
                | self.pw.get_bool(world, 'agentKangaroo')
                | self.pw.get_bool(world, 'agentLemming')
            ),
            if_false=world[:, 6:7, :, :],
            if_true=new_fluid_momentum,
        )

        return world


class BehaviorIce:
    """
    Ice melting. Ice touching water or air turns to water.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return True

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info
        ice_chance = rand_interact
        ice_melting_neighbors = (
            self.pw.get_elem(world, 'empty')
            + self.pw.get_elem(world, 'fire')
            + self.pw.get_elem(world, 'lava')
            + self.pw.get_elem(world, 'water')
        )
        ice_can_melt = conv2d(ice_melting_neighbors, self.pw.neighbor_kernel, padding=1) > 1
        does_turn_water = self.pw.get_bool(world, 'ice') & ice_can_melt & (ice_chance < 0.02)
        world[:] = interp(switch=does_turn_water, if_false=world, if_true=self.pw.elem_vecs['water'])
        return world


class BehaviorWater:
    """
    Water freezing. Water touching 3+ ices can turn to ice.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return True

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info
        ice_chance = rand_element
        water_can_freeze = conv2d(self.pw.get_elem(world, 'ice'), self.pw.neighbor_kernel, padding=1) >= 3
        does_turn_ice = self.pw.get_bool(world, 'water') & water_can_freeze & (ice_chance < 0.05)
        world[:] = interp(switch=does_turn_ice, if_false=world, if_true=self.pw.elem_vecs['ice'])
        return world


class BehaviorFire:
    """
    Fire burning.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return True

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info
        burn_chance = rand_interact
        fire_and_lava = self.pw.get_elem(world, 'fire') + self.pw.get_elem(world, 'lava')
        has_fire_neighbor = conv2d(fire_and_lava, self.pw.neighbor_kernel, padding=1) > 0
        does_burn_wood = self.pw.get_bool(world, 'wood') & (burn_chance < 0.05)
        does_burn_bird = self.pw.get_bool(world, 'agentBird') & (burn_chance < 0.05)
        does_burn_plant = self.pw.get_bool(world, 'plant') & (burn_chance < 0.2)
        does_burn_agent = (
            self.pw.get_bool(world, 'agentFish')
            | self.pw.get_bool(world, 'agentLemming')
            | self.pw.get_bool(world, 'agentKangaroo')
            | self.pw.get_bool(world, 'agentMole')
        ) & (burn_chance < 0.2)
        does_burn_gas = self.pw.get_bool(world, 'gas') & (burn_chance < 0.2)
        does_burn_dust = self.pw.get_bool(world, 'dust')
        does_burn_ice = self.pw.get_bool(world, 'ice') & (burn_chance < 0.2) & has_fire_neighbor
        does_burn = (
            does_burn_wood | does_burn_plant | does_burn_gas | does_burn_dust | does_burn_bird | does_burn_agent
        ) & has_fire_neighbor

        # Velocity for fire
        world[:, 3:5] -= 8 * get_left(does_burn & has_fire_neighbor) * self.pw.left
        world[:, 3:5] -= 8 * get_above(does_burn & has_fire_neighbor) * self.pw.up
        world[:, 3:5] -= 8 * get_below(does_burn & has_fire_neighbor) * self.pw.down
        world[:, 3:5] -= 8 * get_right(does_burn & has_fire_neighbor) * self.pw.right

        world[:, 3:5] -= 30 * get_left(does_burn_dust & has_fire_neighbor) * self.pw.left
        world[:, 3:5] -= 30 * get_above(does_burn_dust & has_fire_neighbor) * self.pw.up
        world[:, 3:5] -= 30 * get_below(does_burn_dust & has_fire_neighbor) * self.pw.down
        world[:, 3:5] -= 30 * get_right(does_burn_dust & has_fire_neighbor) * self.pw.right

        world[:] = interp(switch=does_burn, if_false=world, if_true=self.pw.elem_vecs['fire'])
        world[:] = interp(switch=does_burn_ice, if_false=world, if_true=self.pw.elem_vecs['water'])

        # Fire spread. (Fire+burnable, or Lava)=> creates a probability to spread to air.
        burnables = (
            self.pw.get_elem(world, 'wood')
            + self.pw.get_elem(world, 'plant')
            + self.pw.get_elem(world, 'gas')
            + self.pw.get_elem(world, 'dust')
            + self.pw.get_bool(world, 'agentFish')
            + self.pw.get_bool(world, 'agentBird')
            + self.pw.get_bool(world, 'agentKangaroo')
            + self.pw.get_bool(world, 'agentMole')
            + self.pw.get_bool(world, 'agentLemming')
        )
        fire_with_burnable_neighbor = conv2d(burnables, self.pw.neighbor_kernel, padding=1) * fire_and_lava
        in_fire_range = conv2d(
            fire_with_burnable_neighbor + self.pw.get_elem(world, 'lava'), self.pw.neighbor_kernel, padding=1
        )
        does_burn_empty = self.pw.get_bool(world, 'empty') & (in_fire_range > 0) & (burn_chance < 0.3)
        world[:] = interp(switch=does_burn_empty, if_false=world, if_true=self.pw.elem_vecs['fire'])

        # Fire fading. Fire just has a chance to fade, if not next to a burnable neighbor.
        fire_chance = rand_element
        has_burnable_neighbor = conv2d(burnables, self.pw.neighbor_kernel, padding=1)
        does_fire_turn_empty = self.pw.get_bool(world, 'fire') & (fire_chance < 0.4) & (has_burnable_neighbor == 0)
        world[:] = interp(switch=does_fire_turn_empty, if_false=world, if_true=self.pw.elem_vecs['empty'])

        return world


class BehaviorPlant:
    """
    Plants-growing. If there is water next to plant, and < 4 neighbors, chance to grow there.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return True

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info
        plant_chance = rand_interact
        plant_counts = conv2d(self.pw.get_elem(world, 'plant'), self.pw.neighbor_kernel, padding=1)
        does_plantgrow = self.pw.get_bool(world, 'water') & (plant_chance < 0.05)
        does_plantgrow_plant = does_plantgrow & (plant_counts <= 3) & (plant_counts >= 1)
        does_plantgrow_empty = does_plantgrow & (plant_counts > 3)

        wood_ice_counts = conv2d(
            self.pw.get_elem(world, 'ice') + self.pw.get_elem(world, 'wood'), self.pw.neighbor_kernel, padding=1
        )
        does_plantgrow_plant = does_plantgrow_plant | (
            (wood_ice_counts > 0) & (plant_chance < 0.2) & self.pw.get_bool(world, 'empty') & (plant_counts > 0)
        )

        world[:] = interp2(
            switch_a=does_plantgrow_plant,
            switch_b=does_plantgrow_empty,
            if_false=world,
            if_a=self.pw.elem_vecs['plant'],
            if_b=self.pw.elem_vecs['empty'],
        )
        return world


class BehaviorLava:
    """
    Lava-water interaction. Lava that is touching water turns to stone.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return True

    def forward(self, world, info):
        water_counts = conv2d(self.pw.get_elem(world, 'water'), self.pw.neighbor_kernel, padding=1)
        does_turn_stone = (water_counts > 0) & self.pw.get_bool(world, 'lava')
        world[:] = interp(switch=does_turn_stone, if_false=world, if_true=self.pw.elem_vecs['stone'])

        lava_counts = conv2d(self.pw.get_elem(world, 'lava'), self.pw.neighbor_kernel, padding=1)
        does_turn_stone = (lava_counts > 0) & self.pw.get_bool(world, 'sand')
        world[:] = interp(switch=does_turn_stone, if_false=world, if_true=self.pw.elem_vecs['stone'])
        return world


class BehaviorAcid:
    """
    Acid destroys everything except wall and cloner.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return True

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info
        acid_rand = rand_interact < 0.2
        is_block = ~(
            self.pw.get_bool(world, 'empty')
            | self.pw.get_bool(world, 'wall')
            | self.pw.get_bool(world, 'acid')
            | self.pw.get_bool(world, 'cloner')
            | self.pw.get_bool(world, 'agentSnake')
            | self.pw.get_bool(world, 'gas')
        )
        is_acid = self.pw.get_bool(world, 'acid')
        does_acid_dissapear = (is_acid & acid_rand & get_below(is_block)) | (is_acid & acid_rand & get_above(is_block))
        does_block_dissapear = (is_block & get_above(acid_rand) & get_above(is_acid)) | (
            is_block & get_below(acid_rand) & get_below(is_acid)
        )
        does_dissapear = does_acid_dissapear | does_block_dissapear
        world[:] = interp(switch=does_dissapear, if_false=world, if_true=self.pw.elem_vecs['gas'])
        return world


class BehaviorCloner:
    """
    Cloner keeps track of the first element it touches, and then replaces neighboring empty blocks with that element.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return self.pw.get_bool(world, 'cloner').any().item()

    def forward(self, world, info):
        cloner_assigns = world[:, 6:7, :, :]
        is_not_cloner = ~self.pw.get_bool(world, 'cloner')
        labels = world[:, 0:1]
        for get_dir in [get_below, get_above, get_left, get_right]:
            is_cloner_empty = self.pw.get_bool(world, 'cloner') & ((cloner_assigns == 0) | (cloner_assigns == 13))
            dir_labels = get_dir(labels)
            world[:, 6:7, :, :] = interp2(
                switch_a=is_not_cloner,
                switch_b=is_cloner_empty,
                if_a=cloner_assigns,
                if_b=dir_labels,
                if_false=cloner_assigns,
            )

        # Cloner produce
        cloner_assigns_ids = np.clip(world[:, 6], 0, self.pw.NUM_ELEMENTS - 1).astype(int)
        cloner_assigns_vec = self.pw.elem_vecs_array[cloner_assigns_ids]
        cloner_assigns_vec = np.transpose(cloner_assigns_vec, (0, 3, 1, 2))
        for get_dir in [get_below, get_above, get_left, get_right]:
            cloner_assigns_vec_dir = get_dir(cloner_assigns_vec)
            is_dir_cloner_not_empty = get_dir(
                self.pw.get_bool(world, 'cloner') & ((cloner_assigns != 0) & (cloner_assigns != 13))
            ) & self.pw.get_bool(world, 'empty')
            world[:] = interp(switch=is_dir_cloner_not_empty, if_false=world, if_true=cloner_assigns_vec_dir)
        return world


class BehaviorVelocity:
    """
    Velocity field movement
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return (world[:, 3:5].abs() > 0.9).any().item()

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info
        velocity_field = world[:, 3:5]

        for n in range(2):
            velocity_field_magnitudes = np.linalg.norm(velocity_field, axis=1)[:, None]
            velocity_field_angles_raw = (1 / (2 * np.pi)) * np.arccos(
                velocity_field[:, 1:2] / (velocity_field_magnitudes + 0.001)
            )
            is_y_lessthan_zero = velocity_field[:, 0:1] < 0
            velocity_field_angle = interp(
                switch=is_y_lessthan_zero, if_false=velocity_field_angles_raw, if_true=(1 - velocity_field_angles_raw)
            )
            velocity_field_delta = velocity_field.copy()
            velocity_angle_int = np.remainder(np.floor(velocity_field_angle * 8 + 0.5), 8)
            is_velocity_enough = (velocity_field_magnitudes > (1.0 if n == 0 else 2.0)) & (
                ~self.pw.get_bool(world, 'wall')
            )

            dw = []
            for angle in [0, 1, 2, 3, 4, 5, 6, 7]:
                dw.append(self.pw.direction_func(angle, world))

            swaps = -np.ones((world.shape[0], 1, world.shape[2], world.shape[3]), dtype=self.pw.pw_type)
            for angle in [0, 1, 2, 3, 4, 5, 6, 7]:
                direction_empty = self.pw.get_bool(dw[angle], 'empty')
                direction_swap = self.pw.direction_func(angle, swaps)
                match = (
                    (velocity_angle_int == angle)
                    & is_velocity_enough
                    & (swaps == -1)
                    & (direction_swap == -1)
                    & direction_empty
                )
                opposite_match = self.pw.direction_func((angle + 4) % 8, match)
                swaps = interp_int(match, swaps, angle)
                swaps = interp_int(opposite_match, swaps, (angle + 4) % 8)

            velocity_field_old = velocity_field.copy()
            world[:] = interp_swaps8(swaps, world, dw[0], dw[1], dw[2], dw[3], dw[4], dw[5], dw[6], dw[7])
            world[:, 3:5] = world[:, 3:5] * 0.5 + velocity_field_old * 0.5

        # Velocity field reduction
        velocity_field *= 0.95
        for i in range(1):
            velocity_field[:, 0:1] = (
                conv2d(velocity_field[:, 0:1], self.pw.neighbor_kernel / 18, padding=1) + velocity_field[:, 0:1] * 0.5
            )
            velocity_field[:, 1:2] = (
                conv2d(velocity_field[:, 1:2], self.pw.neighbor_kernel / 18, padding=1) + velocity_field[:, 1:2] * 0.5
            )
        world[:, 3:5] = velocity_field
        return world


class BehaviorFish:
    """
    Fish move randomly.
    IF (Fish & direction) -> Become opposite direction.
    IF in opposite direction is (Fish & direction) -> become Fish.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return self.pw.get_bool(world, 'agentFish').any().item()

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info

        # Small issue here: the order matters because sometimes fish move twice if they roll the correct angle.
        # We could fix it by keeping track of new fish and not allowing them to move.

        for angle in [0, 1, 2, 3]:
            is_gravity = world[:, 2:3] == 1
            is_angle_match = np.floor(rand_movement * 4) == angle
            density = world[:, 1:2]
            is_empty_in_dir = self.pw.direction_func(angle * 2, is_gravity & (density <= 2))
            is_fish = self.pw.get_bool(world, 'agentFish')
            opposite_world = self.pw.direction_func(angle * 2, world)
            does_become_opposite = (
                is_angle_match
                & is_empty_in_dir
                & is_fish
                & self.pw.direction_func(angle * 2, ~is_fish)
                & (rand_interact < 0.2)
            )
            does_become_fish = self.pw.direction_func(((angle * 2) + 4) % 8, does_become_opposite)

            world[:] = interp2(
                switch_a=does_become_fish,
                switch_b=does_become_opposite,
                if_false=world,
                if_a=self.pw.elem_vecs['agentFish'],
                if_b=opposite_world,
            )

        return world


class BehaviorBird:
    """
    Birds have a random velocity, and create velocity in that direction.
    """

    def __init__(self, pw):
        self.pw = pw
        self.obstacle_kernel = np.concatenate(
            [
                np.broadcast_to((np.arange(7) - 3)[None, None, :, None], (1, 1, 7, 7)),
                np.broadcast_to((np.arange(7) - 3)[None, None, None, :], (1, 1, 7, 7)),
            ],
            axis=0,
        ).astype(self.pw.pw_type)
        self.flocking_kernel = np.ones((2, 2, 13, 13), dtype=self.pw.pw_type)
        self.flocking_kernel[1, 0] = 0
        self.flocking_kernel[0, 1] = 0
        self.flocking_kernel[:, :, 3, 3] = 0

    def check_filter(self, world):
        return self.pw.get_bool(world, 'agentBird').any().item()

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info

        is_empty_bird_vel = (world[:, 6:7] == 0) & (world[:, 7:8] == 0)
        is_empty_bird = is_empty_bird_vel & self.pw.get_bool(world, 'agentBird')
        random_dirs = np.concatenate([np.cos(rand_movement * np.pi * 2), np.sin(rand_movement * np.pi * 2)], axis=1)
        bird_vel = world[:, 6:8].copy()
        bird_vel = interp(switch=is_empty_bird, if_false=bird_vel, if_true=random_dirs)

        not_empty = (~self.pw.get_bool(world, 'empty')).astype(self.pw.pw_type)
        vel_delta_obstacle = -conv2d(not_empty, self.obstacle_kernel, padding=3)
        vel_delta_flocking = 1 * conv2d(
            bird_vel * self.pw.get_elem(world, 'agentBird'), self.flocking_kernel, padding=6
        )

        bird_vel += self.pw.get_elem(world, 'agentBird') * (vel_delta_obstacle + vel_delta_flocking)
        bird_vel = normalize(bird_vel + 0.01, axis=1)

        original_68 = world[:, 6:8].copy()
        world[:, 3:5] += self.pw.get_elem(world, 'agentBird') * bird_vel
        world[:, 6:8] = interp(self.pw.get_bool(world, 'agentBird'), original_68, bird_vel)

        return world


class BehaviorKangaroo:
    """
    Kangaroos move left/right and also randomly jump and pick up blocks.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return True

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info

        density = world[:, 1:2]
        fluid_momentum = world[:, 6:7]
        kangaroo_jump_state = world[:, 7:8]
        is_kangaroo = self.pw.get_bool(world, 'agentKangaroo')

        is_kangaroo_jump = is_kangaroo & (rand_element < 0.05) & get_below(density >= 3)
        kangaroo_jump_state = interp(switch=is_kangaroo_jump, if_false=kangaroo_jump_state, if_true=self.pw.one)
        kangaroo_jump_state = interp(
            switch=is_kangaroo, if_false=kangaroo_jump_state, if_true=kangaroo_jump_state - 0.1
        )
        world[:, 7:8] = kangaroo_jump_state

        world[:, 3:5] += (is_kangaroo & (kangaroo_jump_state > 0)) * self.pw.up * 4

        return world


class BehaviorMole:
    """
    Moles burrow through solids.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return self.pw.get_bool(world, 'agentMole').any().item()

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info

        beetle_dir = world[:, 6:7]

        is_beetle = self.pw.get_bool(world, 'agentMole')
        does_beetle_dir_change = is_beetle & (rand_element < 0.1)
        new_beetle_dir = np.floor(rand_movement * 4)
        beetle_dir = interp(switch=does_beetle_dir_change, if_false=beetle_dir, if_true=new_beetle_dir)
        world[:, 6:7] = beetle_dir

        density = world[:, 1:2]
        is_beetle_num = self.pw.get_elem(world, 'agentMole')
        has_supports = conv2d((density >= 3).astype(self.pw.pw_type), self.pw.neighbor_kernel, padding=1)
        world[:, 2:3] = (1 - is_beetle_num) * world[:, 2:3] + is_beetle_num * (has_supports < 2)

        dw = []
        for angle in [0, 1, 2, 3]:
            dw.append(self.pw.direction_func(angle * 2, world))

        for angle in [0, 1, 2, 3]:
            is_angle_match = beetle_dir == angle
            is_solid_in_dir = dw[angle][:, 1:2] >= 3
            is_wall_in_dir = self.pw.get_bool(dw[angle], 'wall')
            is_empty_in_dir = self.pw.get_bool(dw[angle], 'empty')
            is_beetle_in_dor = self.pw.get_bool(dw[angle], 'agentMole')
            does_move_in_dir = ((rand_element < 0.5) & is_solid_in_dir & ~is_wall_in_dir) | (
                (rand_element < 0.1) & ~is_wall_in_dir
            )
            does_become_opposite = is_angle_match & does_move_in_dir & is_beetle & ~is_beetle_in_dor
            does_become_beetle = self.pw.direction_func(((angle * 2) + 4) % 8, does_become_opposite)

            resulting_world = interp(switch=is_solid_in_dir, if_false=dw[angle], if_true=self.pw.elem_vecs['empty'])

            world[:] = interp2(
                switch_a=does_become_beetle,
                switch_b=does_become_opposite,
                if_false=world,
                if_a=dw[(angle + 2) % 4],
                if_b=resulting_world,
            )

        return world


class BehaviorLemming:
    """
    If lemmings run into a block, they move up.
    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return True

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info

        fluid_momentum = world[:, 6:7, :, :]
        for fallLeft in [True, False]:
            fall_dir = fluid_momentum > 0.5
            is_matching_fall = fall_dir if fallLeft else (~fall_dir)
            get_in_dir = get_left if fallLeft else get_right
            get_in_not_dir = get_right if fallLeft else get_left
            is_element = self.pw.get_bool(world, 'agentLemming')

            density = world[:, 1:2]
            density_forward = get_in_dir(density)
            density_above = get_above(density)
            density_forward_above = get_in_dir(get_above(density))

            is_forward_density_higher = (density_forward - density) >= 0
            is_above_density_lower = (density_above - density) < 0
            is_density_forward_above_lower = (density_forward_above - density) < 0

            does_become_above = (
                is_element & is_forward_density_higher & is_above_density_lower & is_density_forward_above_lower
            )
            does_become_below = get_below(does_become_above)

            world_above = get_above(world)
            world_below = get_below(world)
            world[:] = interp2(
                switch_a=does_become_above,
                switch_b=does_become_below,
                if_false=world,
                if_a=world_above,
                if_b=world_below,
            )

        return world


class BehaviorSnake:
    """
    [s1][s2][e]
    [s=does_become_opposite -> become snake_trail]
    [e=does_become_snake -> become swap]
    Snake is working, but we need to prevent turns into itself / turns into the wall.
    How to do so? How about don't turn if there is a wall present or there is more snake present?

    """

    def __init__(self, pw):
        self.pw = pw

    def check_filter(self, world):
        return self.pw.get_bool(world, 'agentSnake').any().item()

    def forward(self, world, info):
        rand_movement, rand_interact, rand_element = info

        snake_dir = world[:, 6:7]
        was_snake = self.pw.get_bool(world, 'agentSnake').copy()
        old_snake_dir = snake_dir.copy()
        old_snake_energy = world[:, 7:8].copy()

        does_become_trail = np.zeros((world.shape[0], 1, world.shape[2], world.shape[3]), dtype=bool)
        does_become_snake = np.zeros((world.shape[0], 1, world.shape[2], world.shape[3]), dtype=bool)
        dir_snake_came_from = np.zeros((world.shape[0], 1, world.shape[2], world.shape[3]), dtype=self.pw.pw_type)
        ones = np.ones((world.shape[0], 1, world.shape[2], world.shape[3]), dtype=self.pw.pw_type)

        does_turn = rand_movement < 0.1

        # For each angle, if it's a snake, become trail.
        # If coming from behind is this angle -> does_become_this_angle = (It becomes a snake. There is a snake before entering the tile.)
        # If random, or infront is a wall, does_turn.
        # Dir_snake_come_from = which angle entering from?
        for angle in [0, 1, 2, 3]:
            is_angle_match = snake_dir == angle
            is_snake_angle = self.pw.get_bool(world, 'agentSnake') & is_angle_match
            does_become_trail = does_become_trail | is_snake_angle
            does_become_this_angle = self.pw.direction_func(((angle * 2) + 4) % 8, is_snake_angle) & ~self.pw.get_bool(
                world, 'wall'
            )
            does_turn = does_turn | (
                self.pw.direction_func(
                    angle * 2, self.pw.get_bool(world, 'wall') | self.pw.get_bool(world, 'agentSnake')
                )
                & does_become_this_angle
            )
            world[:] = interp((does_become_snake & does_become_this_angle), world, self.pw.elem_vecs['empty'])
            does_become_snake = does_become_snake | does_become_this_angle
            dir_snake_came_from = interp(
                switch=does_become_this_angle, if_false=dir_snake_came_from, if_true=ones * angle
            )

            # Perform Swaps
        acid_or_not = interp(rand_element < 0.05, self.pw.elem_vecs['empty'], self.pw.elem_vecs['acid'])
        snake_trail = interp(old_snake_energy > 0, acid_or_not, self.pw.elem_vecs['agentSnake'])
        world[:] = interp(does_become_trail, if_false=world, if_true=snake_trail)
        world[:] = interp(does_become_snake, if_false=world, if_true=self.pw.elem_vecs['agentSnake'])

        # This is where I am about to turn in.
        turned_dir_came_from = (dir_snake_came_from + 1 - 2 * (rand_element < 0.5).astype(int)) % 4
        in_dir = get_in_cardinal_direction(world, turned_dir_came_from * 2)
        # BUT, don't turn if there is a snake or wall in that direction!
        does_turn = does_turn & ~self.pw.get_bool(in_dir, 'agentSnake') & ~self.pw.get_bool(in_dir, 'wall')

        is_snake = self.pw.get_bool(world, 'agentSnake')
        dir_snake_came_from = interp(switch=does_turn, if_false=dir_snake_came_from, if_true=turned_dir_came_from)
        new_snake_dir = interp(switch=(is_snake & ~was_snake), if_false=old_snake_dir, if_true=dir_snake_came_from)

        world[:, 6:7] = interp(is_snake, world[:, 6:7], new_snake_dir)
        world[:, 7:8] = interp(is_snake, world[:, 7:8], old_snake_energy - 0.1)

        return world



from typing import Any

import flax
import jax
import jax.numpy as jnp
import ml_collections
import optax
from utils.encoders import GCEncoder, encoder_modules
from utils.flax_utils import ModuleDict, TrainState, nonpytree_field
from utils.networks import GCActor, GCDiscreteActor


class GCBCAgent(flax.struct.PyTreeNode):
    """Goal-conditioned behavioral cloning (GCBC) agent."""

    rng: Any
    network: Any
    config: Any = nonpytree_field()

    def actor_loss(self, batch, grad_params, rng=None):
        """Compute the BC actor loss."""
        dist = self.network.select('actor')(batch['observations'], batch['actor_goals'], params=grad_params)
        log_prob = dist.log_prob(batch['actions'])

        actor_loss = -log_prob.mean()

        actor_info = {
            'actor_loss': actor_loss,
            'bc_log_prob': log_prob.mean(),
        }
        if not self.config['discrete']:
            actor_info.update(
                {
                    'mse': jnp.mean((dist.mode() - batch['actions']) ** 2),
                    'std': jnp.mean(dist.scale_diag),
                }
            )

        return actor_loss, actor_info

    @jax.jit
    def total_loss(self, batch, grad_params, rng=None):
        """Compute the total loss."""
        info = {}
        rng = rng if rng is not None else self.rng

        rng, actor_rng = jax.random.split(rng)
        actor_loss, actor_info = self.actor_loss(batch, grad_params, actor_rng)
        for k, v in actor_info.items():
            info[f'actor/{k}'] = v

        loss = actor_loss
        return loss, info

    @jax.jit
    def update(self, batch):
        """Update the agent and return a new agent with information dictionary."""
        new_rng, rng = jax.random.split(self.rng)

        def loss_fn(grad_params):
            return self.total_loss(batch, grad_params, rng=rng)

        new_network, info = self.network.apply_loss_fn(loss_fn=loss_fn)

        return self.replace(network=new_network, rng=new_rng), info

    @jax.jit
    def sample_actions(
        self,
        observations,
        goals=None,
        seed=None,
        temperature=1.0,
    ):
        """Sample actions from the actor."""
        dist = self.network.select('actor')(observations, goals, temperature=temperature)
        actions = dist.sample(seed=seed)
        if not self.config['discrete']:
            actions = jnp.clip(actions, -1, 1)
        return actions

    @classmethod
    def create(
        cls,
        seed,
        ex_observations,
        ex_actions,
        config,
    ):
        """Create a new agent.

        Args:
            seed: Random seed.
            ex_observations: Example batch of observations.
            ex_actions: Example batch of actions. In discrete-action MDPs, this should contain the maximum action value.
            config: Configuration dictionary.
        """
        rng = jax.random.PRNGKey(seed)
        rng, init_rng = jax.random.split(rng, 2)

        ex_goals = ex_observations
        if config['discrete']:
            action_dim = ex_actions.max() + 1
        else:
            action_dim = ex_actions.shape[-1]

        # Define encoder.
        encoders = dict()
        if config['encoder'] is not None:
            encoder_module = encoder_modules[config['encoder']]
            encoders['actor'] = GCEncoder(concat_encoder=encoder_module())

        # Define actor network.
        if config['discrete']:
            actor_def = GCDiscreteActor(
                hidden_dims=config['actor_hidden_dims'],
                action_dim=action_dim,
                gc_encoder=encoders.get('actor'),
            )
        else:
            actor_def = GCActor(
                hidden_dims=config['actor_hidden_dims'],
                action_dim=action_dim,
                state_dependent_std=False,
                const_std=config['const_std'],
                gc_encoder=encoders.get('actor'),
            )

        network_info = dict(
            actor=(actor_def, (ex_observations, ex_goals)),
        )
        networks = {k: v[0] for k, v in network_info.items()}
        network_args = {k: v[1] for k, v in network_info.items()}

        network_def = ModuleDict(networks)
        network_tx = optax.adam(learning_rate=config['lr'])
        network_params = network_def.init(init_rng, **network_args)['params']
        network = TrainState.create(network_def, network_params, tx=network_tx)

        return cls(rng, network=network, config=flax.core.FrozenDict(**config))


def get_config():
    config = ml_collections.ConfigDict(
        dict(
            # Agent hyperparameters.
            agent_name='gcbc',  # Agent name.
            lr=3e-4,  # Learning rate.
            batch_size=1024,  # Batch size.
            actor_hidden_dims=(512, 512, 512),  # Actor network hidden dimensions.
            discount=0.99,  # Discount factor (unused by default; can be used for geometric goal sampling in GCDataset).
            const_std=True,  # Whether to use constant standard deviation for the actor.
            discrete=False,  # Whether the action space is discrete.
            encoder=ml_collections.config_dict.placeholder(str),  # Visual encoder name (None, 'impala_small', etc.).
            # Dataset hyperparameters.
            dataset_class='GCDataset',  # Dataset class name.
            value_p_curgoal=0.0,  # Unused (defined for compatibility with GCDataset).
            value_p_trajgoal=1.0,  # Unused (defined for compatibility with GCDataset).
            value_p_randomgoal=0.0,  # Unused (defined for compatibility with GCDataset).
            value_geom_sample=False,  # Unused (defined for compatibility with GCDataset).
            actor_p_curgoal=0.0,  # Probability of using the current state as the actor goal.
            actor_p_trajgoal=1.0,  # Probability of using a future state in the same trajectory as the actor goal.
            actor_p_randomgoal=0.0,  # Probability of using a random state as the actor goal.
            actor_geom_sample=False,  # Whether to use geometric sampling for future actor goals.
            gc_negative=True,  # Unused (defined for compatibility with GCDataset).
            p_aug=0.0,  # Probability of applying image augmentation.
            frame_stack=ml_collections.config_dict.placeholder(int),  # Number of frames to stack.
        )
    )
    return config



import pathlib
from collections import defaultdict

import gymnasium
import numpy as np
from absl import app, flags
from tqdm import trange

import ogbench.powderworld  # noqa
from ogbench.powderworld.behaviors import FillBehavior, LineBehavior, SquareBehavior

FLAGS = flags.FLAGS

flags.DEFINE_integer('seed', 0, 'Random seed.')
flags.DEFINE_string('env_name', 'powderworld-v0', 'Environment name.')
flags.DEFINE_string('dataset_type', 'play', 'Dataset type.')
flags.DEFINE_string('save_path', None, 'Save path.')
flags.DEFINE_integer('num_episodes', 1000, 'Number of episodes.')
flags.DEFINE_integer('max_episode_steps', 1001, 'Maximum number of steps in an episode.')
flags.DEFINE_float('p_random_action', 0.5, 'Probability of selecting a random action.')


def main(_):
    assert FLAGS.dataset_type in ['play']

    # Initialize environment.
    env = gymnasium.make(
        FLAGS.env_name,
        mode='data_collection',
        max_episode_steps=FLAGS.max_episode_steps,
    )
    env.reset()

    # Initialize agents.
    agents = [
        FillBehavior(env=env),
        LineBehavior(env=env),
        SquareBehavior(env=env),
    ]
    probs = np.array([1, 3, 3])  # Agent selection probabilities.
    probs = probs / probs.sum()

    # Collect data.
    dataset = defaultdict(list)
    total_steps = 0
    total_train_steps = 0
    num_train_episodes = FLAGS.num_episodes
    num_val_episodes = FLAGS.num_episodes // 10
    for ep_idx in trange(num_train_episodes + num_val_episodes):
        ob, info = env.reset()
        agent = np.random.choice(agents, p=probs)
        agent.reset(ob, info)

        done = False
        step = 0

        action_step = 0  # Action cycle counter (0, 1, 2).
        while not done:
            if action_step == 0:
                # Select an action every 3 steps.
                if np.random.rand() < FLAGS.p_random_action:
                    # Sample a random action.
                    semantic_action = env.unwrapped.sample_semantic_action()
                else:
                    # Get an action from the agent.
                    semantic_action = agent.select_action(ob, info)
            action = env.unwrapped.semantic_action_to_action(*semantic_action)
            next_ob, reward, terminated, truncated, info = env.step(action)
            done = terminated or truncated

            if agent.done and FLAGS.dataset_type == 'play':
                agent = np.random.choice(agents, p=probs)
                agent.reset(ob, info)

            dataset['observations'].append(ob)
            dataset['actions'].append(action)
            dataset['terminals'].append(done)

            ob = next_ob
            step += 1
            action_step = (action_step + 1) % 3

        total_steps += step
        if ep_idx < num_train_episodes:
            total_train_steps += step

    print('Total steps:', total_steps)

    train_path = FLAGS.save_path
    val_path = FLAGS.save_path.replace('.npz', '-val.npz')
    pathlib.Path(train_path).parent.mkdir(parents=True, exist_ok=True)

    # Split the dataset into training and validation sets.
    train_dataset = {}
    val_dataset = {}
    for k, v in dataset.items():
        if 'observations' in k and v[0].dtype == np.uint8:
            dtype = np.uint8
        elif 'actions':
            dtype = np.int32
        elif k == 'terminals':
            dtype = bool
        else:
            dtype = np.float32
        train_dataset[k] = np.array(v[:total_train_steps], dtype=dtype)
        val_dataset[k] = np.array(v[total_train_steps:], dtype=dtype)

    for path, dataset in [(train_path, train_dataset), (val_path, val_dataset)]:
        np.savez_compressed(path, **dataset)


if __name__ == '__main__':
    app.run(main)



import numpy as np
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter1d

from ogbench.manipspace import lie


class PlanOracle:
    """Non-Markovian oracle that follows a pre-computed plan.

    It first generates a plan by interpolating the keyframes of the task and adds temporally correlated noise. Then, it
    computes the actions by computing the difference between the current state and the next state in the plan.
    """

    def __init__(self, env, segment_dt=0.4, noise=0.1, noise_smoothing=0.5):
        """Initialize the oracle.

        Args:
            env: Environment.
            segment_dt: Default duration of each segment between keyframes in the plan.
            noise: Noise level to add to the plan.
            noise_smoothing: Noise smoothing level.
        """
        self._env = env
        self._env_dt = self._env.unwrapped._control_timestep
        self._dt = segment_dt
        self._noise = noise
        self._noise_smoothing = noise_smoothing

        self._done = False
        self._t_init = None
        self._t_max = None
        self._plan = None

    def above(self, pose, z):
        return (
            lie.SE3.from_rotation_and_translation(
                rotation=lie.SO3.identity(),
                translation=np.array([0.0, 0.0, z]),
            )
            @ pose
        )

    def to_pose(self, pos, yaw):
        return lie.SE3.from_rotation_and_translation(
            rotation=lie.SO3.from_z_radians(yaw),
            translation=pos,
        )

    def get_yaw(self, pose):
        yaw = pose.rotation().compute_yaw_radians()
        if yaw < 0.0:
            return yaw + 2 * np.pi
        return yaw

    def shortest_yaw(self, eff_yaw, obj_yaw, translation, n=4):
        """Find the symmetry-aware shortest yaw angle to the object."""
        symmetries = np.array([i * 2 * np.pi / n + obj_yaw for i in range(-n, n + 1)])
        d = np.argmin(np.abs(eff_yaw - symmetries))
        return lie.SE3.from_rotation_and_translation(
            rotation=lie.SO3.from_z_radians(symmetries[d]),
            translation=translation,
        )

    def compute_plan(self, times, poses, grasps):
        # Interpolate grasps.
        grasp_interp = interp1d(times, grasps, kind='linear', axis=0, assume_sorted=True)

        # Interpolate poses.
        xyzs = [p.translation() for p in poses]
        xyz_interp = interp1d(times, xyzs, kind='linear', axis=0, assume_sorted=True)

        # Interpolate orientations.
        quats = [p.rotation() for p in poses]

        def quat_interp(t):
            s = np.searchsorted(times, t, side='right') - 1
            interp_time = (t - times[s]) / (times[s + 1] - times[s])
            interp_time = np.clip(interp_time, 0.0, 1.0)
            return lie.interpolate(quats[s], quats[s + 1], interp_time)

        # Generate the plan.
        plan = []
        t = 0.0
        while t < self._t_max:
            action = np.zeros(5)
            action[:3] = xyz_interp(t)
            action[3] = quat_interp(t).compute_yaw_radians()
            action[4] = grasp_interp(t)
            plan.append(action)
            t += self._env_dt

        plan = np.array(plan)

        # Add temporally correlated noise to the plan.
        if self._noise > 0:
            noise = np.random.normal(0, 1, size=(len(plan), 5)) * np.array([0.05, 0.05, 0.05, 0.3, 1.0]) * self._noise
            noise = gaussian_filter1d(noise, axis=0, sigma=self._noise_smoothing)
            plan += noise

        return plan

    @property
    def done(self):
        return self._done

    def reset(self, ob, info):
        pass

    def select_action(self, ob, info):
        # Find the current plan index.
        cur_plan_idx = int((info['time'][0] - self._t_init + 1e-7) // self._env_dt)
        if cur_plan_idx >= len(self._plan) - 1:
            cur_plan_idx = len(self._plan) - 1
            self._done = True

        # Compute the difference between the current state and the current plan.
        ab_action = self._plan[cur_plan_idx]
        action = np.zeros(5)
        action[:3] = ab_action[:3] - info['proprio/effector_pos']
        action[3] = ab_action[3] - info['proprio/effector_yaw'][0]
        action[4] = ab_action[4] - info['proprio/gripper_opening'][0]
        action = self._env.unwrapped.normalize_action(action)

        return action



import numpy as np


class MarkovOracle:
    """Markovian oracle for manipulation tasks."""

    def __init__(self, env, min_norm=0.4):
        """Initialize the oracle.

        Args:
            env: Environment.
            min_norm: Minimum norm for the relative position. Setting it to a non-zero value can help the agent to learn
                more robust policies.
        """
        self._env = env
        self._min_norm = min_norm
        self._debug = False  # Set to True to print debug information.
        self._done = False

        if self._debug:
            np.set_printoptions(suppress=True)

    def shape_diff(self, diff):
        """Shape the difference vector to have a minimum norm."""
        diff_norm = np.linalg.norm(diff)
        if diff_norm >= self._min_norm:
            return diff
        else:
            return diff / (diff_norm + 1e-6) * self._min_norm

    def shortest_yaw(self, eff_yaw, obj_yaw, n=4):
        """Find the symmetry-aware shortest yaw angle to the object."""
        symmetries = np.array([i * 2 * np.pi / n + obj_yaw for i in range(-n, n + 1)])
        d = np.argmin(np.abs(eff_yaw - symmetries))
        return symmetries[d]

    def print_phase(self, phase):
        """Print the current phase."""
        if self._debug:
            print(f'Phase {phase:50}', end=' ')

    @property
    def done(self):
        return self._done

    def reset(self, ob, info):
        pass

    def select_action(self, ob, info):
        pass



import os
import urllib.request

import gymnasium
import numpy as np
from tqdm import tqdm

from ogbench.relabel_utils import add_oracle_reps, relabel_dataset

DEFAULT_DATASET_DIR = '~/.ogbench/data'
DATASET_URL = 'https://rail.eecs.berkeley.edu/datasets/ogbench'


def load_dataset(dataset_path, ob_dtype=np.float32, action_dtype=np.float32, compact_dataset=False, add_info=False):
    """Load OGBench dataset.

    Args:
        dataset_path: Path to the dataset file.
        ob_dtype: dtype for observations.
        action_dtype: dtype for actions.
        compact_dataset: Whether to return a compact dataset (True, without 'next_observations') or a regular dataset
            (False, with 'next_observations').
        add_info: Whether to add observation information ('qpos', 'qvel', and 'button_states') to the dataset.

    Returns:
        Dictionary containing the dataset. The dictionary contains the following keys: 'observations', 'actions',
        'terminals', and 'next_observations' (if `compact_dataset` is False) or 'valids' (if `compact_dataset` is True).
        If `add_info` is True, the dictionary may also contain additional keys for observation information.
    """
    file = np.load(dataset_path)

    dataset = dict()
    for k in ['observations', 'actions', 'terminals']:
        if k == 'observations':
            dtype = ob_dtype
        elif k == 'actions':
            dtype = action_dtype
        else:
            dtype = np.float32
        dataset[k] = file[k][...].astype(dtype, copy=False)

    if add_info:
        # Read observation information.
        info_keys = []
        for k in ['qpos', 'qvel', 'button_states']:
            if k in file:
                dataset[k] = file[k][...]
                info_keys.append(k)

    # Example:
    # Assume each trajectory has length 4, and (s0, a0, s1), (s1, a1, s2), (s2, a2, s3), (s3, a3, s4) are transition
    # tuples. Note that (s4, a4, s0) is *not* a valid transition tuple, and a4 does not have a corresponding next state.
    # At this point, `dataset` loaded from the file has the following structure:
    #                  |<--- traj 1 --->|  |<--- traj 2 --->|  ...
    # -------------------------------------------------------------
    # 'observations': [s0, s1, s2, s3, s4, s0, s1, s2, s3, s4, ...]
    # 'actions'     : [a0, a1, a2, a3, a4, a0, a1, a2, a3, a4, ...]
    # 'terminals'   : [ 0,  0,  0,  0,  1,  0,  0,  0,  0,  1, ...]

    if compact_dataset:
        # Compact dataset: We need to invalidate the last state of each trajectory so that we can safely get
        # `next_observations[t]` by using `observations[t + 1]`.
        # Our goal is to have the following structure:
        #                  |<--- traj 1 --->|  |<--- traj 2 --->|  ...
        # -------------------------------------------------------------
        # 'observations': [s0, s1, s2, s3, s4, s0, s1, s2, s3, s4, ...]
        # 'actions'     : [a0, a1, a2, a3, a4, a0, a1, a2, a3, a4, ...]
        # 'terminals'   : [ 0,  0,  0,  1,  1,  0,  0,  0,  1,  1, ...]
        # 'valids'      : [ 1,  1,  1,  1,  0,  1,  1,  1,  1,  0, ...]

        dataset['valids'] = 1.0 - dataset['terminals']
        new_terminals = np.concatenate([dataset['terminals'][1:], [1.0]])
        dataset['terminals'] = np.minimum(dataset['terminals'] + new_terminals, 1.0).astype(np.float32)
    else:
        # Regular dataset: Generate `next_observations` by shifting `observations`.
        # Our goal is to have the following structure:
        #                       |<- traj 1 ->|  |<- traj 2 ->|  ...
        # ----------------------------------------------------------
        # 'observations'     : [s0, s1, s2, s3, s0, s1, s2, s3, ...]
        # 'actions'          : [a0, a1, a2, a3, a0, a1, a2, a3, ...]
        # 'next_observations': [s1, s2, s3, s4, s1, s2, s3, s4, ...]
        # 'terminals'        : [ 0,  0,  0,  1,  0,  0,  0,  1, ...]

        ob_mask = (1.0 - dataset['terminals']).astype(bool)
        next_ob_mask = np.concatenate([[False], ob_mask[:-1]])
        dataset['next_observations'] = dataset['observations'][next_ob_mask]
        dataset['observations'] = dataset['observations'][ob_mask]
        dataset['actions'] = dataset['actions'][ob_mask]
        new_terminals = np.concatenate([dataset['terminals'][1:], [1.0]])
        dataset['terminals'] = new_terminals[ob_mask].astype(np.float32)

        if add_info:
            for k in info_keys:
                dataset[k] = dataset[k][ob_mask]

    return dataset


def download_datasets(dataset_names, dataset_dir=DEFAULT_DATASET_DIR):
    """Download OGBench datasets.

    Args:
        dataset_names: List of dataset names to download.
        dataset_dir: Directory to save the datasets.
    """
    # Make dataset directory.
    dataset_dir = os.path.expanduser(dataset_dir)
    os.makedirs(dataset_dir, exist_ok=True)

    # Download datasets.
    dataset_file_names = []
    for dataset_name in dataset_names:
        dataset_file_names.append(f'{dataset_name}.npz')
        dataset_file_names.append(f'{dataset_name}-val.npz')
    for dataset_file_name in dataset_file_names:
        dataset_file_path = os.path.join(dataset_dir, dataset_file_name)
        if not os.path.exists(dataset_file_path):
            dataset_url = f'{DATASET_URL}/{dataset_file_name}'
            print('Downloading dataset from:', dataset_url)
            response = urllib.request.urlopen(dataset_url)
            tmp_dataset_file_path = f'{dataset_file_path}.tmp'
            with tqdm.wrapattr(
                open(tmp_dataset_file_path, 'wb'),
                'write',
                miniters=1,
                desc=dataset_url.split('/')[-1],
                total=getattr(response, 'length', None),
            ) as file:
                for chunk in response:
                    file.write(chunk)
            os.rename(tmp_dataset_file_path, dataset_file_path)


def make_env_and_datasets(
    dataset_name,
    dataset_dir=DEFAULT_DATASET_DIR,
    dataset_path=None,
    compact_dataset=False,
    env_only=False,
    dataset_only=False,
    cur_env=None,
    add_info=False,
    **env_kwargs,
):
    """Make OGBench environment and load datasets.

    Args:
        dataset_name: Dataset name.
        dataset_dir: Directory to save/load the datasets.
        dataset_path: (Optional) Path to the dataset file.
        compact_dataset: Whether to return a compact dataset (True, without 'next_observations') or a regular dataset
            (False, with 'next_observations').
        env_only: Whether to return only the environment.
        dataset_only: Whether to return only the datasets.
        cur_env: Current environment (only used when `dataset_only` is True).
        add_info: Whether to add observation information ('qpos', 'qvel', and 'button_states') to the datasets.
        **env_kwargs: Keyword arguments to pass to the environment.
    """
    # Make environment.
    splits = dataset_name.split('-')
    dataset_add_info = add_info
    env = cur_env
    if 'singletask' in splits:
        # Single-task environment.
        pos = splits.index('singletask')
        env_name = '-'.join(splits[: pos - 1] + splits[pos:])  # Remove the dataset type.
        if not dataset_only:
            env = gymnasium.make(env_name, **env_kwargs)
        dataset_name = '-'.join(splits[:pos] + splits[-1:])  # Remove the words 'singletask' and 'task\d' (if exists).
        dataset_add_info = True
    elif 'oraclerep' in splits:
        # Environment with oracle goal representations.
        env_name = '-'.join(splits[:-3] + splits[-1:])  # Remove the dataset type and the word 'oraclerep'.
        if not dataset_only:
            env = gymnasium.make(env_name, use_oracle_rep=True, **env_kwargs)
        dataset_name = '-'.join(splits[:-2] + splits[-1:])  # Remove the word 'oraclerep'.
        dataset_add_info = True
    else:
        # Original, goal-conditioned environment.
        env_name = '-'.join(splits[:-2] + splits[-1:])  # Remove the dataset type.
        if not dataset_only:
            env = gymnasium.make(env_name, **env_kwargs)

    if env_only:
        return env

    # Load datasets.
    if dataset_path is None:
        dataset_dir = os.path.expanduser(dataset_dir)
        download_datasets([dataset_name], dataset_dir)
        train_dataset_path = os.path.join(dataset_dir, f'{dataset_name}.npz')
        val_dataset_path = os.path.join(dataset_dir, f'{dataset_name}-val.npz')
    else:
        train_dataset_path = dataset_path
        val_dataset_path = dataset_path.replace('.npz', '-val.npz')

    ob_dtype = np.uint8 if ('visual' in env_name or 'powderworld' in env_name) else np.float32
    action_dtype = np.int32 if 'powderworld' in env_name else np.float32
    train_dataset = load_dataset(
        train_dataset_path,
        ob_dtype=ob_dtype,
        action_dtype=action_dtype,
        compact_dataset=compact_dataset,
        add_info=dataset_add_info,
    )
    val_dataset = load_dataset(
        val_dataset_path,
        ob_dtype=ob_dtype,
        action_dtype=action_dtype,
        compact_dataset=compact_dataset,
        add_info=dataset_add_info,
    )

    if 'singletask' in splits:
        # Add reward information to the datasets.
        relabel_dataset(env_name, env, train_dataset)
        relabel_dataset(env_name, env, val_dataset)

    if 'oraclerep' in splits:
        # Add oracle goal representations to the datasets.
        add_oracle_reps(env_name, env, train_dataset)
        add_oracle_reps(env_name, env, val_dataset)

    if not add_info:
        # Remove information keys.
        for k in ['qpos', 'qvel', 'button_states']:
            if k in train_dataset:
                del train_dataset[k]
            if k in val_dataset:
                del val_dataset[k]

    if dataset_only:
        return train_dataset, val_dataset
    else:
        return env, train_dataset, val_dataset



import glob
import json
import pathlib
from collections import defaultdict

import gymnasium
import numpy as np
from absl import app, flags
from agents import SACAgent
from tqdm import trange
from utils.evaluation import supply_rng
from utils.flax_utils import restore_agent

import ogbench.locomaze  # noqa

FLAGS = flags.FLAGS

flags.DEFINE_integer('seed', 0, 'Random seed.')
flags.DEFINE_string('env_name', 'antsoccer-arena-v0', 'Environment name.')
flags.DEFINE_string('dataset_type', 'navigate', 'Dataset type.')
flags.DEFINE_string('loco_restore_path', 'experts/ant', 'Locomotion agent restore path.')
flags.DEFINE_integer('loco_restore_epoch', 400000, 'Locomotion agent restore epoch.')
flags.DEFINE_string('ball_restore_path', 'experts/antball', 'Ball agent restore path.')
flags.DEFINE_integer('ball_restore_epoch', 12000000, 'Ball agent restore epoch.')
flags.DEFINE_string('save_path', None, 'Save path.')
flags.DEFINE_float('noise', 0.2, 'Gaussian action noise level.')
flags.DEFINE_integer('num_episodes', 1000, 'Number of episodes.')
flags.DEFINE_integer('max_episode_steps', 1001, 'Maximum number of steps in an episode.')


def load_agent(restore_path, restore_epoch, ob_dim, action_dim):
    """Initialize and load a SAC agent from a given path."""
    # Load agent config.
    candidates = glob.glob(restore_path)
    assert len(candidates) == 1, f'Found {len(candidates)} candidates: {candidates}'

    with open(candidates[0] + '/flags.json', 'r') as f:
        agent_config = json.load(f)['agent']

    # Load agent.
    agent = SACAgent.create(
        FLAGS.seed,
        np.zeros(ob_dim),
        np.zeros(action_dim),
        agent_config,
    )
    agent = restore_agent(agent, restore_path, restore_epoch)

    return agent


def main(_):
    assert FLAGS.dataset_type in ['navigate', 'stitch']
    # 'navigate': Repeatedly navigate to the ball and then to a goal in a single episode.
    # 'stitch': Either only navigate or only dribble the ball to a goal in a single episode.

    # Initialize environment.
    env = gymnasium.make(
        FLAGS.env_name,
        terminate_at_goal=False,
        max_episode_steps=FLAGS.max_episode_steps,
    )
    ob_dim = env.observation_space.shape[0]
    action_dim = env.action_space.shape[0]

    # Initialize oracle agent.
    loco_agent = load_agent(FLAGS.loco_restore_path, FLAGS.loco_restore_epoch, ob_dim, action_dim)
    ball_agent = load_agent(FLAGS.ball_restore_path, FLAGS.ball_restore_epoch, ob_dim, action_dim)
    loco_actor_fn = supply_rng(loco_agent.sample_actions, rng=loco_agent.rng)
    ball_actor_fn = supply_rng(ball_agent.sample_actions, rng=ball_agent.rng)

    def get_agent_action(ob, goal_xy):
        """Get an action for the agent to navigate to the goal."""
        if 'arena' not in FLAGS.env_name:
            # In the actual maze environment, replace the goal with the oracle subgoal.
            goal_xy, _ = env.unwrapped.get_oracle_subgoal(ob[:2], goal_xy)
        goal_dir = goal_xy - ob[:2]
        goal_dir = goal_dir / (np.linalg.norm(goal_dir) + 1e-6)
        # Concatenate the agent's joint positions (excluding the x-y position), joint velocities, and goal direction.
        agent_ob = np.concatenate([ob[2:15], ob[22:36], goal_dir])
        action = loco_actor_fn(agent_ob, temperature=0)
        return action

    def get_ball_action(ob, ball_xy, goal_xy):
        """Get an action for the agent to dribble the ball to the goal."""
        if 'arena' in FLAGS.env_name:
            if np.linalg.norm(goal_xy - ball_xy) > 10:
                # If the ball is too far from the goal, set a virtual goal 10 units away from the ball. This is because
                # the ball agent is not trained to dribble the ball to the goal that is too far away.
                goal_xy = ball_xy + 10 * (goal_xy - ball_xy) / np.linalg.norm(goal_xy - ball_xy)
        else:
            # In the actual maze environment, replace the goal with the oracle subgoal.
            goal_xy, _ = env.unwrapped.get_oracle_subgoal(ball_xy, goal_xy)
        # Concatenate the agent and ball's joint positions (excluding their x-y positions), their joint velocities, and
        # the relative positions of the ball and the goal.
        agent_ob = np.concatenate([ob[2:15], ob[17:], ball_xy - agent_xy, goal_xy - ball_xy])
        action = ball_actor_fn(agent_ob, temperature=0)
        return action

    # Store all empty cells.
    all_cells = []
    maze_map = env.unwrapped.maze_map
    for i in range(maze_map.shape[0]):
        for j in range(maze_map.shape[1]):
            if maze_map[i, j] == 0:
                all_cells.append((i, j))

    # Collect data.
    dataset = defaultdict(list)
    total_steps = 0
    total_train_steps = 0
    num_train_episodes = FLAGS.num_episodes
    num_val_episodes = FLAGS.num_episodes // 10
    for ep_idx in trange(num_train_episodes + num_val_episodes):
        if FLAGS.dataset_type == 'navigate':
            # Sample random initial positions for the agent, the ball, and the goal.
            agent_init_idx, ball_init_idx, goal_idx = np.random.choice(len(all_cells), 3, replace=False)
            agent_init_ij = all_cells[agent_init_idx]
            ball_init_ij = all_cells[ball_init_idx]
            goal_ij = all_cells[goal_idx]
        elif FLAGS.dataset_type == 'stitch':
            # Randomly choose between the 'navigate' and 'dribble' modes.
            cur_mode = 'navigate' if np.random.randint(2) == 0 else 'dribble'

            # Sample random initial positions for the agent, the ball, and the goal. In the 'dribble' mode, the ball
            # always starts at the agent's position.
            agent_init_idx, ball_init_idx, goal_idx = np.random.choice(len(all_cells), 3, replace=False)
            agent_init_ij = all_cells[agent_init_idx]
            ball_init_ij = all_cells[ball_init_idx] if cur_mode == 'navigate' else agent_init_ij
            goal_ij = all_cells[goal_idx]
        else:
            raise ValueError(f'Unsupported dataset_type: {FLAGS.dataset_type}')

        ob, _ = env.reset(
            options=dict(task_info=dict(agent_init_ij=agent_init_ij, ball_init_ij=ball_init_ij, goal_ij=goal_ij))
        )

        done = False
        step = 0

        virtual_agent_goal_xy = None  # Virtual goal for the agent to move to when stuck.

        while not done:
            agent_xy, ball_xy = env.unwrapped.get_agent_ball_xy()
            agent_xy, ball_xy = np.array(agent_xy), np.array(ball_xy)
            goal_xy = np.array(env.unwrapped.cur_goal_xy)

            if FLAGS.dataset_type == 'navigate':
                if virtual_agent_goal_xy is None:
                    if np.linalg.norm(agent_xy - ball_xy) > 2:
                        # If the agent is far from the ball, move to the ball.
                        action = get_agent_action(ob, ball_xy)
                    else:
                        # If the agent is close to the ball, dribble the ball to the goal.
                        action = get_ball_action(ob, ball_xy, goal_xy)
                else:
                    # When virtual_agent_goal_xy is set, move to the virtual goal.
                    action = get_agent_action(ob, virtual_agent_goal_xy)
            elif FLAGS.dataset_type == 'stitch':
                if cur_mode == 'navigate':
                    # Navigate to the goal.
                    action = get_agent_action(ob, goal_xy)
                else:
                    # Dribble the ball to the goal.
                    action = get_ball_action(ob, ball_xy, goal_xy)

            # Add Gaussian noise to the action.
            action = action + np.random.normal(0, FLAGS.noise, action.shape)
            action = np.clip(action, -1, 1)

            next_ob, reward, terminated, truncated, info = env.step(action)
            done = terminated or truncated
            success = info['success']

            if virtual_agent_goal_xy is not None and np.linalg.norm(virtual_agent_goal_xy - next_ob[:2]) <= 0.5:
                # If the agent reaches the virtual goal, clear it.
                virtual_agent_goal_xy = None

            if FLAGS.dataset_type == 'navigate':
                if success:
                    # Sample a new goal state when the current goal is reached.
                    goal_ij = all_cells[np.random.randint(len(all_cells))]
                    env.unwrapped.set_goal(goal_ij)

                # Determine whether the agent is stuck.
                if (
                    step > 150
                    and virtual_agent_goal_xy is None
                    and np.linalg.norm(np.array(dataset['observations'][-150:])[:, :2] - next_ob[:2], axis=1).max() <= 2
                ):
                    # When the agent is stuck for 150 steps, set a virtual goal to move to a random cell.
                    virtual_agent_goal_ij = all_cells[np.random.randint(len(all_cells))]
                    virtual_agent_goal_xy = np.array(env.unwrapped.ij_to_xy(virtual_agent_goal_ij))

            dataset['observations'].append(ob)
            dataset['actions'].append(action)
            dataset['terminals'].append(done)
            dataset['qpos'].append(info['prev_qpos'])
            dataset['qvel'].append(info['prev_qvel'])

            ob = next_ob
            step += 1

        total_steps += step
        if ep_idx < num_train_episodes:
            total_train_steps += step

    print('Total steps:', total_steps)

    train_path = FLAGS.save_path
    val_path = FLAGS.save_path.replace('.npz', '-val.npz')
    pathlib.Path(train_path).parent.mkdir(parents=True, exist_ok=True)

    # Split the dataset into training and validation sets.
    train_dataset = {
        k: np.array(v[:total_train_steps], dtype=np.float32 if k != 'terminals' else bool) for k, v in dataset.items()
    }
    val_dataset = {
        k: np.array(v[total_train_steps:], dtype=np.float32 if k != 'terminals' else bool) for k, v in dataset.items()
    }

    for path, dataset in [(train_path, train_dataset), (val_path, val_dataset)]:
        np.savez_compressed(path, **dataset)


if __name__ == '__main__':
    app.run(main)



from ogbench.manipspace.lie.se3 import SE3
from ogbench.manipspace.lie.so3 import SO3
from ogbench.manipspace.lie.utils import get_epsilon, interpolate, mat2quat, skew

__all__ = (
    'SE3',
    'SO3',
    'get_epsilon',
    'interpolate',
    'mat2quat',
    'skew',
)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import numpy as np

from ogbench.manipspace.lie.so3 import SO3
from ogbench.manipspace.lie.utils import get_epsilon, skew

_IDENTITY_WXYZ_XYZ = np.array([1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], dtype=np.float64)


@dataclass(frozen=True)
class SE3:
    """Special Euclidean group for proper rigid transforms in 3D.

    Internal parameterization is (qw, qx, qy, qz, x, y, z).
    Tangent parameterization is (vx, vy, vz, omega_x, omega_y, omega_z).
    """

    wxyz_xyz: np.ndarray
    matrix_dim: int = 4
    parameters_dim: int = 7
    tangent_dim: int = 6
    space_dim: int = 3

    def __repr__(self) -> str:
        quat = np.round(self.wxyz_xyz[:4], 5)
        xyz = np.round(self.wxyz_xyz[4:], 5)
        return f'{self.__class__.__name__}(wxyz={quat}, xyz={xyz})'

    @staticmethod
    def identity() -> SE3:
        return SE3(wxyz_xyz=_IDENTITY_WXYZ_XYZ)

    @staticmethod
    def from_rotation_and_translation(
        rotation: SO3,
        translation: np.ndarray,
    ) -> SE3:
        assert translation.shape == (SE3.space_dim,)
        return SE3(wxyz_xyz=np.concatenate([rotation.wxyz, translation]))

    @staticmethod
    def from_matrix(matrix: np.ndarray) -> SE3:
        assert matrix.shape == (SE3.matrix_dim, SE3.matrix_dim)
        return SE3.from_rotation_and_translation(
            rotation=SO3.from_matrix(matrix[:3, :3]),
            translation=matrix[:3, 3],
        )

    @staticmethod
    def sample_uniform() -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=SO3.sample_uniform(),
            translation=np.random.uniform(-1.0, 1.0, size=(SE3.space_dim,)),
        )

    def rotation(self) -> SO3:
        return SO3(wxyz=self.wxyz_xyz[:4])

    def translation(self) -> np.ndarray:
        return self.wxyz_xyz[4:]

    def as_matrix(self) -> np.ndarray:
        hmat = np.eye(self.matrix_dim, dtype=np.float64)
        hmat[:3, :3] = self.rotation().as_matrix()
        hmat[:3, 3] = self.translation()
        return hmat

    @staticmethod
    def exp(tangent: np.ndarray) -> SE3:
        assert tangent.shape == (SE3.tangent_dim,)
        rotation = SO3.exp(tangent[3:])
        theta_squared = tangent[3:] @ tangent[3:]
        use_taylor = theta_squared < get_epsilon(theta_squared.dtype)
        theta_squared_safe = 1.0 if use_taylor else theta_squared
        theta_safe = np.sqrt(theta_squared_safe)
        skew_omega = skew(tangent[3:])
        if use_taylor:
            V = rotation.as_matrix()
        else:
            V = (
                np.eye(3, dtype=np.float64)
                + (1.0 - np.cos(theta_safe)) / (theta_squared_safe) * skew_omega
                + (theta_safe - np.sin(theta_safe)) / (theta_squared_safe * theta_safe) * (skew_omega @ skew_omega)
            )
        return SE3.from_rotation_and_translation(
            rotation=rotation,
            translation=V @ tangent[:3],
        )

    def log(self) -> np.ndarray:
        omega = self.rotation().log()
        theta_squared = omega @ omega
        use_taylor = theta_squared < get_epsilon(theta_squared.dtype)
        skew_omega = skew(omega)
        theta_squared_safe = 1.0 if use_taylor else theta_squared
        theta_safe = np.sqrt(theta_squared_safe)
        half_theta_safe = 0.5 * theta_safe
        if use_taylor:
            V_inv = np.eye(3, dtype=np.float64) - 0.5 * skew_omega + (skew_omega @ skew_omega) / 12.0
        else:
            V_inv = (
                np.eye(3, dtype=np.float64)
                - 0.5 * skew_omega
                + (1.0 - theta_safe * np.cos(half_theta_safe) / (2.0 * np.sin(half_theta_safe)))
                / theta_squared_safe
                * (skew_omega @ skew_omega)
            )
        return np.concatenate([V_inv @ self.translation(), omega])

    def adjoint(self) -> np.ndarray:
        R = self.rotation().as_matrix()
        return np.block(
            [
                [R, np.zeros((3, 3), dtype=np.float64)],
                [skew(self.translation()) @ R, R],
            ]
        )

    def inverse(self) -> SE3:
        R_inv = self.rotation().inverse()
        return SE3.from_rotation_and_translation(
            rotation=R_inv,
            translation=-(R_inv @ self.translation()),
        )

    def normalize(self) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=self.rotation().normalize(),
            translation=self.translation(),
        )

    def apply(self, target: np.ndarray) -> np.ndarray:
        assert target.shape == (SE3.space_dim,)
        return self.rotation() @ target + self.translation()

    def multiply(self, other: SE3) -> SE3:
        return SE3.from_rotation_and_translation(
            rotation=self.rotation() @ other.rotation(),
            translation=(self.rotation() @ other.translation()) + self.translation(),
        )

    def __matmul__(self, other: Any) -> Any:
        if isinstance(other, np.ndarray):
            return self.apply(target=other)
        elif isinstance(other, SE3):
            return self.multiply(other=other)
        else:
            raise ValueError(f'Unsupported argument type for @ operator: {type(other)}')



import numpy as np

from ogbench.manipspace.oracles.markov.markov_oracle import MarkovOracle


class ButtonMarkovOracle(MarkovOracle):
    def __init__(self, max_step=100, gripper_always_closed=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._max_step = max_step
        self._gripper_always_closed = gripper_always_closed

    def reset(self, ob, info):
        self._done = False
        self._step = 0
        self._final_pos = np.random.uniform(*self._env.unwrapped._arm_sampling_bounds)
        self._final_yaw = np.random.uniform(-np.pi, np.pi)

    def select_action(self, ob, info):
        effector_pos = info['proprio/effector_pos']
        effector_yaw = info['proprio/effector_yaw'][0]

        target_button = info['privileged/target_button']
        button_target_top_pos = info['privileged/target_button_top_pos'] + np.array([0, 0, 0.06])
        button_target_bottom_pos = info['privileged/target_button_top_pos'] - np.array([0, 0, 0.022])
        button_state = info[f'privileged/button_{target_button}_state']
        target_state = info['privileged/target_button_state']

        above_threshold = 0.16
        above = effector_pos[2] > above_threshold
        xy_aligned = np.linalg.norm(button_target_top_pos[:2] - effector_pos[:2]) <= 0.04
        target_achieved = button_state == target_state
        final_pos_aligned = np.linalg.norm(self._final_pos - effector_pos) <= 0.04

        gain_pos = 5
        gain_yaw = 3
        action = np.zeros(5)
        if not target_achieved:
            if not xy_aligned:
                self.print_phase('1: Move above the button')
                action = np.zeros(5)
                diff = button_target_top_pos - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[4] = 1
            else:
                self.print_phase('2: Press the button')
                action = np.zeros(5)
                diff = button_target_bottom_pos - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[4] = 1
        else:
            if not above:
                self.print_phase('3: Release the button')
                diff = (
                    np.array([button_target_top_pos[0], button_target_top_pos[1], above_threshold * 2]) - effector_pos
                )
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (self._final_yaw - effector_yaw) * gain_yaw
                action[4] = 1 if self._gripper_always_closed else -1
            else:
                self.print_phase('4: Move to the final position')
                diff = self._final_pos - effector_pos
                diff = self.shape_diff(diff)
                action[:3] = diff[:3] * gain_pos
                action[3] = (self._final_yaw - effector_yaw) * gain_yaw
                action[4] = 1 if self._gripper_always_closed else -1

            if final_pos_aligned:
                self._done = True

        action = np.clip(action, -1, 1)
        if self._debug:
            print(action)

        self._step += 1
        if self._step == self._max_step:
            self._done = True

        return action



import dataclasses
from functools import partial
from typing import Any

import jax
import jax.numpy as jnp
import numpy as np
from flax.core.frozen_dict import FrozenDict


def get_size(data):
    """Return the size of the dataset."""
    sizes = jax.tree_util.tree_map(lambda arr: len(arr), data)
    return max(jax.tree_util.tree_leaves(sizes))


@partial(jax.jit, static_argnames=('padding',))
def random_crop(img, crop_from, padding):
    """Randomly crop an image.

    Args:
        img: Image to crop.
        crop_from: Coordinates to crop from.
        padding: Padding size.
    """
    padded_img = jnp.pad(img, ((padding, padding), (padding, padding), (0, 0)), mode='edge')
    return jax.lax.dynamic_slice(padded_img, crop_from, img.shape)


@partial(jax.jit, static_argnames=('padding',))
def batched_random_crop(imgs, crop_froms, padding):
    """Batched version of random_crop."""
    return jax.vmap(random_crop, (0, 0, None))(imgs, crop_froms, padding)


class Dataset(FrozenDict):
    """Dataset class.

    This class supports both regular datasets (i.e., storing both observations and next_observations) and
    compact datasets (i.e., storing only observations). It assumes 'observations' is always present in the keys. If
    'next_observations' is not present, it will be inferred from 'observations' by shifting the indices by 1. In this
    case, set 'valids' appropriately to mask out the last state of each trajectory.
    """

    @classmethod
    def create(cls, freeze=True, **fields):
        """Create a dataset from the fields.

        Args:
            freeze: Whether to freeze the arrays.
            **fields: Keys and values of the dataset.
        """
        data = fields
        assert 'observations' in data
        if freeze:
            jax.tree_util.tree_map(lambda arr: arr.setflags(write=False), data)
        return cls(data)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.size = get_size(self._dict)
        if 'valids' in self._dict:
            (self.valid_idxs,) = np.nonzero(self['valids'] > 0)

    def get_random_idxs(self, num_idxs):
        """Return `num_idxs` random indices."""
        if 'valids' in self._dict:
            return self.valid_idxs[np.random.randint(len(self.valid_idxs), size=num_idxs)]
        else:
            return np.random.randint(self.size, size=num_idxs)

    def sample(self, batch_size, idxs=None):
        """Sample a batch of transitions."""
        if idxs is None:
            idxs = self.get_random_idxs(batch_size)
        return self.get_subset(idxs)

    def get_subset(self, idxs):
        """Return a subset of the dataset given the indices."""
        result = jax.tree_util.tree_map(lambda arr: arr[idxs], self._dict)
        if 'next_observations' not in result:
            result['next_observations'] = self._dict['observations'][np.minimum(idxs + 1, self.size - 1)]
        return result


class ReplayBuffer(Dataset):
    """Replay buffer class.

    This class extends Dataset to support adding transitions.
    """

    @classmethod
    def create(cls, transition, size):
        """Create a replay buffer from the example transition.

        Args:
            transition: Example transition (dict).
            size: Size of the replay buffer.
        """

        def create_buffer(example):
            example = np.array(example)
            return np.zeros((size, *example.shape), dtype=example.dtype)

        buffer_dict = jax.tree_util.tree_map(create_buffer, transition)
        return cls(buffer_dict)

    @classmethod
    def create_from_initial_dataset(cls, init_dataset, size):
        """Create a replay buffer from the initial dataset.

        Args:
            init_dataset: Initial dataset.
            size: Size of the replay buffer.
        """

        def create_buffer(init_buffer):
            buffer = np.zeros((size, *init_buffer.shape[1:]), dtype=init_buffer.dtype)
            buffer[: len(init_buffer)] = init_buffer
            return buffer

        buffer_dict = jax.tree_util.tree_map(create_buffer, init_dataset)
        dataset = cls(buffer_dict)
        dataset.size = dataset.pointer = get_size(init_dataset)
        return dataset

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.max_size = get_size(self._dict)
        self.size = 0
        self.pointer = 0

    def add_transition(self, transition):
        """Add a transition to the replay buffer."""

        def set_idx(buffer, new_element):
            buffer[self.pointer] = new_element

        jax.tree_util.tree_map(set_idx, self._dict, transition)
        self.pointer = (self.pointer + 1) % self.max_size
        self.size = max(self.pointer, self.size)

    def clear(self):
        """Clear the replay buffer."""
        self.size = self.pointer = 0


@dataclasses.dataclass
class GCDataset:
    """Dataset class for goal-conditioned RL.

    This class provides a method to sample a batch of transitions with goals (value_goals and actor_goals) from the
    dataset. The goals are sampled from the current state, future states in the same trajectory, and random states.
    It also supports frame stacking and random-cropping image augmentation.

    It reads the following keys from the config:
    - discount: Discount factor for geometric sampling.
    - value_p_curgoal: Probability of using the current state as the value goal.
    - value_p_trajgoal: Probability of using a future state in the same trajectory as the value goal.
    - value_p_randomgoal: Probability of using a random state as the value goal.
    - value_geom_sample: Whether to use geometric sampling for future value goals.
    - actor_p_curgoal: Probability of using the current state as the actor goal.
    - actor_p_trajgoal: Probability of using a future state in the same trajectory as the actor goal.
    - actor_p_randomgoal: Probability of using a random state as the actor goal.
    - actor_geom_sample: Whether to use geometric sampling for future actor goals.
    - gc_negative: Whether to use '0 if s == g else -1' (True) or '1 if s == g else 0' (False) as the reward.
    - p_aug: Probability of applying image augmentation.
    - frame_stack: Number of frames to stack.

    Attributes:
        dataset: Dataset object.
        config: Configuration dictionary.
        preprocess_frame_stack: Whether to preprocess frame stacks. If False, frame stacks are computed on-the-fly. This
            saves memory but may slow down training.
    """

    dataset: Dataset
    config: Any
    preprocess_frame_stack: bool = True

    def __post_init__(self):
        self.size = self.dataset.size

        # Pre-compute trajectory boundaries.
        (self.terminal_locs,) = np.nonzero(self.dataset['terminals'] > 0)
        self.initial_locs = np.concatenate([[0], self.terminal_locs[:-1] + 1])
        assert self.terminal_locs[-1] == self.size - 1

        # Assert probabilities sum to 1.
        assert np.isclose(
            self.config['value_p_curgoal'] + self.config['value_p_trajgoal'] + self.config['value_p_randomgoal'], 1.0
        )
        assert np.isclose(
            self.config['actor_p_curgoal'] + self.config['actor_p_trajgoal'] + self.config['actor_p_randomgoal'], 1.0
        )

        if self.config['frame_stack'] is not None:
            # Only support compact (observation-only) datasets.
            assert 'next_observations' not in self.dataset
            if self.preprocess_frame_stack:
                stacked_observations = self.get_stacked_observations(np.arange(self.size))
                self.dataset = Dataset(self.dataset.copy(dict(observations=stacked_observations)))

    def sample(self, batch_size, idxs=None, evaluation=False):
        """Sample a batch of transitions with goals.

        This method samples a batch of transitions with goals (value_goals and actor_goals) from the dataset. They are
        stored in the keys 'value_goals' and 'actor_goals', respectively. It also computes the 'rewards' and 'masks'
        based on the indices of the goals.

        Args:
            batch_size: Batch size.
            idxs: Indices of the transitions to sample. If None, random indices are sampled.
            evaluation: Whether to sample for evaluation. If True, image augmentation is not applied.
        """
        if idxs is None:
            idxs = self.dataset.get_random_idxs(batch_size)

        batch = self.dataset.sample(batch_size, idxs)
        if self.config['frame_stack'] is not None:
            batch['observations'] = self.get_observations(idxs)
            batch['next_observations'] = self.get_observations(idxs + 1)

        value_goal_idxs = self.sample_goals(
            idxs,
            self.config['value_p_curgoal'],
            self.config['value_p_trajgoal'],
            self.config['value_p_randomgoal'],
            self.config['value_geom_sample'],
        )
        actor_goal_idxs = self.sample_goals(
            idxs,
            self.config['actor_p_curgoal'],
            self.config['actor_p_trajgoal'],
            self.config['actor_p_randomgoal'],
            self.config['actor_geom_sample'],
        )

        batch['value_goals'] = self.get_observations(value_goal_idxs)
        batch['actor_goals'] = self.get_observations(actor_goal_idxs)
        successes = (idxs == value_goal_idxs).astype(float)
        batch['masks'] = 1.0 - successes
        batch['rewards'] = successes - (1.0 if self.config['gc_negative'] else 0.0)

        if self.config['p_aug'] is not None and not evaluation:
            if np.random.rand() < self.config['p_aug']:
                self.augment(batch, ['observations', 'next_observations', 'value_goals', 'actor_goals'])

        return batch

    def sample_goals(self, idxs, p_curgoal, p_trajgoal, p_randomgoal, geom_sample):
        """Sample goals for the given indices."""
        batch_size = len(idxs)

        # Random goals.
        random_goal_idxs = self.dataset.get_random_idxs(batch_size)

        # Goals from the same trajectory (excluding the current state, unless it is the final state).
        final_state_idxs = self.terminal_locs[np.searchsorted(self.terminal_locs, idxs)]
        if geom_sample:
            # Geometric sampling.
            offsets = np.random.geometric(p=1 - self.config['discount'], size=batch_size)  # in [1, inf)
            traj_goal_idxs = np.minimum(idxs + offsets, final_state_idxs)
        else:
            # Uniform sampling.
            distances = np.random.rand(batch_size)  # in [0, 1)
            traj_goal_idxs = np.round(
                (np.minimum(idxs + 1, final_state_idxs) * distances + final_state_idxs * (1 - distances))
            ).astype(int)
        if p_curgoal == 1.0:
            goal_idxs = idxs
        else:
            goal_idxs = np.where(
                np.random.rand(batch_size) < p_trajgoal / (1.0 - p_curgoal), traj_goal_idxs, random_goal_idxs
            )

            # Goals at the current state.
            goal_idxs = np.where(np.random.rand(batch_size) < p_curgoal, idxs, goal_idxs)

        return goal_idxs

    def augment(self, batch, keys):
        """Apply image augmentation to the given keys."""
        padding = 3
        batch_size = len(batch[keys[0]])
        crop_froms = np.random.randint(0, 2 * padding + 1, (batch_size, 2))
        crop_froms = np.concatenate([crop_froms, np.zeros((batch_size, 1), dtype=np.int64)], axis=1)
        for key in keys:
            batch[key] = jax.tree_util.tree_map(
                lambda arr: np.array(batched_random_crop(arr, crop_froms, padding)) if len(arr.shape) == 4 else arr,
                batch[key],
            )

    def get_observations(self, idxs):
        """Return the observations for the given indices."""
        if self.config['frame_stack'] is None or self.preprocess_frame_stack:
            return jax.tree_util.tree_map(lambda arr: arr[idxs], self.dataset['observations'])
        else:
            return self.get_stacked_observations(idxs)

    def get_stacked_observations(self, idxs):
        """Return the frame-stacked observations for the given indices."""
        initial_state_idxs = self.initial_locs[np.searchsorted(self.initial_locs, idxs, side='right') - 1]
        rets = []
        for i in reversed(range(self.config['frame_stack'])):
            cur_idxs = np.maximum(idxs - i, initial_state_idxs)
            rets.append(jax.tree_util.tree_map(lambda arr: arr[cur_idxs], self.dataset['observations']))
        return jax.tree_util.tree_map(lambda *args: np.concatenate(args, axis=-1), *rets)


@dataclasses.dataclass
class HGCDataset(GCDataset):
    """Dataset class for hierarchical goal-conditioned RL.

    This class extends GCDataset to support high-level actor goals and prediction targets. It reads the following
    additional key from the config:
    - subgoal_steps: Subgoal steps (i.e., the number of steps to reach the low-level goal).
    """

    def sample(self, batch_size, idxs=None, evaluation=False):
        """Sample a batch of transitions with goals.

        This method samples a batch of transitions with goals from the dataset. The goals are stored in the keys
        'value_goals', 'low_actor_goals', 'high_actor_goals', and 'high_actor_targets'. It also computes the 'rewards'
        and 'masks' based on the indices of the goals.

        Args:
            batch_size: Batch size.
            idxs: Indices of the transitions to sample. If None, random indices are sampled.
            evaluation: Whether to sample for evaluation. If True, image augmentation is not applied.
        """
        if idxs is None:
            idxs = self.dataset.get_random_idxs(batch_size)

        batch = self.dataset.sample(batch_size, idxs)
        if self.config['frame_stack'] is not None:
            batch['observations'] = self.get_observations(idxs)
            batch['next_observations'] = self.get_observations(idxs + 1)

        # Sample value goals.
        value_goal_idxs = self.sample_goals(
            idxs,
            self.config['value_p_curgoal'],
            self.config['value_p_trajgoal'],
            self.config['value_p_randomgoal'],
            self.config['value_geom_sample'],
        )
        batch['value_goals'] = self.get_observations(value_goal_idxs)

        successes = (idxs == value_goal_idxs).astype(float)
        batch['masks'] = 1.0 - successes
        batch['rewards'] = successes - (1.0 if self.config['gc_negative'] else 0.0)

        # Set low-level actor goals.
        final_state_idxs = self.terminal_locs[np.searchsorted(self.terminal_locs, idxs)]
        low_goal_idxs = np.minimum(idxs + self.config['subgoal_steps'], final_state_idxs)
        batch['low_actor_goals'] = self.get_observations(low_goal_idxs)

        # Sample high-level actor goals and set prediction targets.
        # High-level future goals.
        if self.config['actor_geom_sample']:
            # Geometric sampling.
            offsets = np.random.geometric(p=1 - self.config['discount'], size=batch_size)  # in [1, inf)
            high_traj_goal_idxs = np.minimum(idxs + offsets, final_state_idxs)
        else:
            # Uniform sampling.
            distances = np.random.rand(batch_size)  # in [0, 1)
            high_traj_goal_idxs = np.round(
                (np.minimum(idxs + 1, final_state_idxs) * distances + final_state_idxs * (1 - distances))
            ).astype(int)
        high_traj_target_idxs = np.minimum(idxs + self.config['subgoal_steps'], high_traj_goal_idxs)

        # High-level random goals.
        high_random_goal_idxs = self.dataset.get_random_idxs(batch_size)
        high_random_target_idxs = np.minimum(idxs + self.config['subgoal_steps'], final_state_idxs)

        # Pick between high-level future goals and random goals.
        pick_random = np.random.rand(batch_size) < self.config['actor_p_randomgoal']
        high_goal_idxs = np.where(pick_random, high_random_goal_idxs, high_traj_goal_idxs)
        high_target_idxs = np.where(pick_random, high_random_target_idxs, high_traj_target_idxs)

        batch['high_actor_goals'] = self.get_observations(high_goal_idxs)
        batch['high_actor_targets'] = self.get_observations(high_target_idxs)

        if self.config['p_aug'] is not None and not evaluation:
            if np.random.rand() < self.config['p_aug']:
                self.augment(
                    batch,
                    [
                        'observations',
                        'next_observations',
                        'value_goals',
                        'low_actor_goals',
                        'high_actor_goals',
                        'high_actor_targets',
                    ],
                )

        return batch



"""OGBench: Benchmarking Offline Goal-Conditioned RL"""

import ogbench.locomaze
import ogbench.manipspace
import ogbench.powderworld
from ogbench.utils import download_datasets, load_dataset, make_env_and_datasets

__all__ = (
    'locomaze',
    'manipspace',
    'powderworld',
    'download_datasets',
    'load_dataset',
    'make_env_and_datasets',
)



import contextlib
import os

import gymnasium
import mujoco
import numpy as np
from gymnasium import utils
from gymnasium.envs.mujoco import MujocoEnv
from gymnasium.spaces import Box


class HumanoidEnv(MujocoEnv, utils.EzPickle):
    """DMC Humanoid environment.

    Several methods are reimplemented to remove the dependency on the `dm_control` package. It is supposed to work
    identically to the original Humanoid environment.
    """

    xml_file = os.path.join(os.path.dirname(__file__), 'assets', 'humanoid.xml')
    metadata = {
        'render_modes': ['human', 'rgb_array', 'depth_array'],
        'render_fps': 40,
    }
    if gymnasium.__version__ >= '1.1.0':
        metadata['render_modes'] += ['rgbd_tuple']

    def __init__(
        self,
        xml_file=None,
        render_mode='rgb_array',
        width=200,
        height=200,
        **kwargs,
    ):
        """Initialize the Humanoid environment.

        Args:
            xml_file: Path to the XML description (optional).
            render_mode: Rendering mode.
            width: Width of the rendered image.
            height: Height of the rendered image.
            **kwargs: Additional keyword arguments.
        """
        if xml_file is None:
            xml_file = self.xml_file
        utils.EzPickle.__init__(
            self,
            xml_file,
            **kwargs,
        )

        observation_space = Box(low=-np.inf, high=np.inf, shape=(69,), dtype=np.float64)

        MujocoEnv.__init__(
            self,
            xml_file,
            frame_skip=5,
            observation_space=observation_space,
            render_mode=render_mode,
            width=width,
            height=height,
            **kwargs,
        )

    def step(self, action):
        prev_qpos = self.data.qpos.copy()
        prev_qvel = self.data.qvel.copy()

        self.do_simulation(action, self.frame_skip)

        qpos = self.data.qpos.copy()
        qvel = self.data.qvel.copy()

        observation = self.get_ob()

        if self.render_mode == 'human':
            self.render()

        return (
            observation,
            0.0,
            False,
            False,
            {
                'xy': self.get_xy(),
                'prev_qpos': prev_qpos,
                'prev_qvel': prev_qvel,
                'qpos': qpos,
                'qvel': qvel,
            },
        )

    def _step_mujoco_simulation(self, ctrl, n_frames):
        self.data.ctrl[:] = ctrl

        # DMC-style stepping (see Page 6 of https://arxiv.org/abs/2006.12983).
        if self.model.opt.integrator != mujoco.mjtIntegrator.mjINT_RK4.value:
            mujoco.mj_step2(self.model, self.data)
            if n_frames > 1:
                mujoco.mj_step(self.model, self.data, n_frames - 1)
        else:
            mujoco.mj_step(self.model, self.data, n_frames)

        mujoco.mj_step1(self.model, self.data)

    def get_ob(self):
        xy = self.data.qpos[:2]
        joint_angles = self.data.qpos[7:]  # Skip the 7 DoFs of the free root joint.
        head_height = self.data.xpos[2, 2]  # ['head', 'z']
        torso_frame = self.data.xmat[1].reshape(3, 3)  # ['torso']
        torso_pos = self.data.xpos[1]  # ['torso']
        positions = []
        for idx in [16, 10, 13, 7]:  # ['left_hand', 'left_foot', 'right_hand', 'right_foot']
            torso_to_limb = self.data.xpos[idx] - torso_pos
            positions.append(torso_to_limb.dot(torso_frame))
        extremities = np.hstack(positions)
        torso_vertical_orientation = self.data.xmat[1, [6, 7, 8]]  # ['torso', ['zx', 'zy', 'zz']]
        center_of_mass_velocity = self.data.sensordata[0:3]  # ['torso_subtreelinvel']
        velocity = self.data.qvel

        return np.concatenate(
            [
                xy,
                joint_angles,
                [head_height],
                extremities,
                torso_vertical_orientation,
                center_of_mass_velocity,
                velocity,
            ]
        )

    @contextlib.contextmanager
    def disable(self, *flags):
        old_bitmask = self.model.opt.disableflags
        new_bitmask = old_bitmask
        for flag in flags:
            if isinstance(flag, str):
                field_name = 'mjDSBL_' + flag.upper()
                flag = getattr(mujoco.mjtDisableBit, field_name)
            elif isinstance(flag, int):
                flag = mujoco.mjtDisableBit(flag)
            new_bitmask |= flag.value
        self.model.opt.disableflags = new_bitmask
        try:
            yield
        finally:
            self.model.opt.disableflags = old_bitmask

    def reset_model(self):
        penetrating = True
        while penetrating:
            quat = self.np_random.uniform(size=4)
            quat /= np.linalg.norm(quat)
            self.data.qpos[3:7] = quat
            self.data.qvel = 0.1 * self.np_random.standard_normal(self.model.nv)

            for joint_id in range(1, self.model.njnt):
                range_min, range_max = self.model.jnt_range[joint_id]
                self.data.qpos[6 + joint_id] = self.np_random.uniform(range_min, range_max)

            with self.disable('actuation'):
                mujoco.mj_forward(self.model, self.data)
            penetrating = self.data.ncon > 0

        observation = self.get_ob()

        return observation

    def get_xy(self):
        return self.data.qpos[:2].copy()

    def set_xy(self, xy):
        qpos = self.data.qpos.copy()
        qvel = self.data.qvel.copy()
        qpos[:2] = xy
        self.set_state(qpos, qvel)



import functools
import glob
import os
import pickle
from typing import Any, Dict, Mapping, Sequence

import flax
import flax.linen as nn
import jax
import jax.numpy as jnp
import optax

nonpytree_field = functools.partial(flax.struct.field, pytree_node=False)


class ModuleDict(nn.Module):
    """A dictionary of modules.

    This allows sharing parameters between modules and provides a convenient way to access them.

    Attributes:
        modules: Dictionary of modules.
    """

    modules: Dict[str, nn.Module]

    @nn.compact
    def __call__(self, *args, name=None, **kwargs):
        """Forward pass.

        For initialization, call with `name=None` and provide the arguments for each module in `kwargs`.
        Otherwise, call with `name=<module_name>` and provide the arguments for that module.
        """
        if name is None:
            if kwargs.keys() != self.modules.keys():
                raise ValueError(
                    f'When `name` is not specified, kwargs must contain the arguments for each module. '
                    f'Got kwargs keys {kwargs.keys()} but module keys {self.modules.keys()}'
                )
            out = {}
            for key, value in kwargs.items():
                if isinstance(value, Mapping):
                    out[key] = self.modules[key](**value)
                elif isinstance(value, Sequence):
                    out[key] = self.modules[key](*value)
                else:
                    out[key] = self.modules[key](value)
            return out

        return self.modules[name](*args, **kwargs)


class TrainState(flax.struct.PyTreeNode):
    """Custom train state for models.

    Attributes:
        step: Counter to keep track of the training steps. It is incremented by 1 after each `apply_gradients` call.
        apply_fn: Apply function of the model.
        model_def: Model definition.
        params: Parameters of the model.
        tx: optax optimizer.
        opt_state: Optimizer state.
    """

    step: int
    apply_fn: Any = nonpytree_field()
    model_def: Any = nonpytree_field()
    params: Any
    tx: Any = nonpytree_field()
    opt_state: Any

    @classmethod
    def create(cls, model_def, params, tx=None, **kwargs):
        """Create a new train state."""
        if tx is not None:
            opt_state = tx.init(params)
        else:
            opt_state = None

        return cls(
            step=1,
            apply_fn=model_def.apply,
            model_def=model_def,
            params=params,
            tx=tx,
            opt_state=opt_state,
            **kwargs,
        )

    def __call__(self, *args, params=None, method=None, **kwargs):
        """Forward pass.

        When `params` is not provided, it uses the stored parameters.

        The typical use case is to set `params` to `None` when you want to *stop* the gradients, and to pass the current
        traced parameters when you want to flow the gradients. In other words, the default behavior is to stop the
        gradients, and you need to explicitly provide the parameters to flow the gradients.

        Args:
            *args: Arguments to pass to the model.
            params: Parameters to use for the forward pass. If `None`, it uses the stored parameters, without flowing
                the gradients.
            method: Method to call in the model. If `None`, it uses the default `apply` method.
            **kwargs: Keyword arguments to pass to the model.
        """
        if params is None:
            params = self.params
        variables = {'params': params}
        if method is not None:
            method_name = getattr(self.model_def, method)
        else:
            method_name = None

        return self.apply_fn(variables, *args, method=method_name, **kwargs)

    def select(self, name):
        """Helper function to select a module from a `ModuleDict`."""
        return functools.partial(self, name=name)

    def apply_gradients(self, grads, **kwargs):
        """Apply the gradients and return the updated state."""
        updates, new_opt_state = self.tx.update(grads, self.opt_state, self.params)
        new_params = optax.apply_updates(self.params, updates)

        return self.replace(
            step=self.step + 1,
            params=new_params,
            opt_state=new_opt_state,
            **kwargs,
        )

    def apply_loss_fn(self, loss_fn):
        """Apply the loss function and return the updated state and info.

        It additionally computes the gradient statistics and adds them to the dictionary.
        """
        grads, info = jax.grad(loss_fn, has_aux=True)(self.params)

        grad_max = jax.tree_util.tree_map(jnp.max, grads)
        grad_min = jax.tree_util.tree_map(jnp.min, grads)
        grad_norm = jax.tree_util.tree_map(jnp.linalg.norm, grads)

        grad_max_flat = jnp.concatenate([jnp.reshape(x, -1) for x in jax.tree_util.tree_leaves(grad_max)], axis=0)
        grad_min_flat = jnp.concatenate([jnp.reshape(x, -1) for x in jax.tree_util.tree_leaves(grad_min)], axis=0)
        grad_norm_flat = jnp.concatenate([jnp.reshape(x, -1) for x in jax.tree_util.tree_leaves(grad_norm)], axis=0)

        final_grad_max = jnp.max(grad_max_flat)
        final_grad_min = jnp.min(grad_min_flat)
        final_grad_norm = jnp.linalg.norm(grad_norm_flat, ord=1)

        info.update(
            {
                'grad/max': final_grad_max,
                'grad/min': final_grad_min,
                'grad/norm': final_grad_norm,
            }
        )

        return self.apply_gradients(grads=grads), info


def save_agent(agent, save_dir, epoch):
    """Save the agent to a file.

    Args:
        agent: Agent.
        save_dir: Directory to save the agent.
        epoch: Epoch number.
    """

    save_dict = dict(
        agent=flax.serialization.to_state_dict(agent),
    )
    save_path = os.path.join(save_dir, f'params_{epoch}.pkl')
    with open(save_path, 'wb') as f:
        pickle.dump(save_dict, f)

    print(f'Saved to {save_path}')


def restore_agent(agent, restore_path, restore_epoch):
    """Restore the agent from a file.

    Args:
        agent: Agent.
        restore_path: Path to the directory containing the saved agent.
        restore_epoch: Epoch number.
    """
    candidates = glob.glob(restore_path)

    assert len(candidates) == 1, f'Found {len(candidates)} candidates: {candidates}'

    restore_path = candidates[0] + f'/params_{restore_epoch}.pkl'

    with open(restore_path, 'rb') as f:
        load_dict = pickle.load(f)

    agent = flax.serialization.from_state_dict(agent, load_dict['agent'])

    print(f'Restored from {restore_path}')

    return agent



import matplotlib
import numpy as np
from matplotlib import figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas


def get_2d_colors(points, min_point, max_point):
    """Get colors corresponding to 2-D points."""
    points = np.array(points)
    min_point = np.array(min_point)
    max_point = np.array(max_point)

    colors = (points - min_point) / (max_point - min_point)
    colors = np.hstack((colors, (2 - np.sum(colors, axis=1, keepdims=True)) / 2))
    colors = np.clip(colors, 0, 1)
    colors = np.c_[colors, np.full(len(colors), 0.8)]

    return colors


def visualize_trajs(env_name, trajs):
    """Visualize x-y trajectories in locomotion environments.

    It reads 'xy' and 'direction' from the 'info' field of the trajectories.
    """
    matplotlib.use('Agg')

    fig = figure.Figure(tight_layout=True)
    canvas = FigureCanvas(fig)
    if 'xy' in trajs[0]['info'][0]:
        ax = fig.add_subplot()

        max_xy = 0.0
        for traj in trajs:
            xy = np.array([info['xy'] for info in traj['info']])
            direction = np.array([info['direction'] for info in traj['info']])
            color = get_2d_colors(direction, [-1, -1], [1, 1])
            for i in range(len(xy) - 1):
                ax.plot(xy[i : i + 2, 0], xy[i : i + 2, 1], color=color[i], linewidth=0.7)
            max_xy = max(max_xy, np.abs(xy).max() * 1.2)

        plot_axis = [-max_xy, max_xy, -max_xy, max_xy]
        ax.axis(plot_axis)
        ax.set_aspect('equal')
    else:
        return None

    fig.tight_layout()
    canvas.draw()
    out_image = np.frombuffer(canvas.tostring_rgb(), dtype='uint8')
    out_image = out_image.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    return out_image



import mujoco
import numpy as np


def get_epsilon(dtype: np.dtype) -> float:
    return {
        np.dtype('float32'): 1e-5,
        np.dtype('float64'): 1e-10,
    }[dtype]


def skew(x: np.ndarray) -> np.ndarray:
    assert x.shape == (3,)
    wx, wy, wz = x
    return np.array(
        [
            [0.0, -wz, wy],
            [wz, 0.0, -wx],
            [-wy, wx, 0.0],
        ]
    )


def mat2quat(mat: np.ndarray):
    """Convert a MuJoCo matrix (9,) to a quaternion (4,)."""
    assert mat.shape == (9,)
    quat = np.empty(4, dtype=np.float64)
    mujoco.mju_mat2Quat(quat, mat)
    return quat


def interpolate(p0, p1, alpha=0.5):
    """Interpolate between two points on a manifold."""
    assert 0.0 <= alpha <= 1.0
    exp_func = getattr(type(p0), 'exp')
    return p0 @ exp_func(alpha * (p0.inverse() @ p1).log())



from gymnasium.envs.registration import register

register(
    id='online-ant-v0',
    entry_point='ogbench.online_locomotion.ant:AntEnv',
    max_episode_steps=1000,
)
register(
    id='online-antball-v0',
    entry_point='ogbench.online_locomotion.ant_ball:AntBallEnv',
    max_episode_steps=200,
)
register(
    id='online-humanoid-v0',
    entry_point='ogbench.online_locomotion.humanoid:HumanoidEnv',
    max_episode_steps=1000,
)



import numpy as np

from ogbench.manipspace import lie
from ogbench.manipspace.oracles.plan.plan_oracle import PlanOracle


class CubePlanOracle(PlanOracle):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    def compute_keyframes(self, plan_input):
        # Poses.
        poses = {}

        # Pick.
        block_initial = self.shortest_yaw(
            eff_yaw=self.get_yaw(plan_input['effector_initial']),
            obj_yaw=self.get_yaw(plan_input['block_initial']),
            translation=plan_input['block_initial'].translation(),
        )
        poses['initial'] = plan_input['effector_initial']
        poses['pick'] = self.above(block_initial, 0.1 + np.random.uniform(0, 0.1))
        poses['pick_start'] = block_initial
        poses['pick_end'] = block_initial
        poses['postpick'] = poses['pick']

        # Place.
        block_goal = self.shortest_yaw(
            eff_yaw=self.get_yaw(poses['postpick']),
            obj_yaw=self.get_yaw(plan_input['block_goal']),
            translation=plan_input['block_goal'].translation(),
        )
        poses['place'] = self.above(block_goal, 0.1 + np.random.uniform(0, 0.1))
        poses['place_start'] = block_goal
        poses['place_end'] = block_goal
        poses['postplace'] = poses['place']
        poses['final'] = plan_input['effector_goal']

        # Clearance.
        midway = lie.interpolate(poses['postpick'], poses['place'])
        poses['clearance'] = lie.SE3.from_rotation_and_translation(
            rotation=midway.rotation(),
            translation=np.array([*midway.translation()[:2], poses['initial'].translation()[-1]])
            + np.random.uniform([-0.1, -0.1, 0], [0.1, 0.1, 0.2]),
        )

        # Times.
        times = {}
        times['initial'] = 0.0
        times['pick'] = times['initial'] + self._dt
        times['pick_start'] = times['pick'] + self._dt * 1.5
        times['pick_end'] = times['pick_start'] + self._dt
        times['postpick'] = times['pick_end'] + self._dt
        times['clearance'] = times['postpick'] + self._dt
        times['place'] = times['clearance'] + self._dt
        times['place_start'] = times['place'] + self._dt * 1.5
        times['place_end'] = times['place_start'] + self._dt
        times['postplace'] = times['place_end'] + self._dt
        times['final'] = times['postplace'] + self._dt
        for time in times.keys():
            if time != 'initial':
                times[time] += np.random.uniform(-1, 1) * self._dt * 0.2

        # Grasps.
        g = 0.0
        grasps = {}
        for name in times.keys():
            if name in {'pick_end', 'place_end'}:
                g = 1.0 - g
            grasps[name] = g

        return times, poses, grasps

    def reset(self, ob, info):
        target_block = info['privileged/target_block']
        plan_input = {
            'effector_initial': self.to_pose(
                pos=info['proprio/effector_pos'],
                yaw=info['proprio/effector_yaw'][0],
            ),
            'effector_goal': self.to_pose(
                pos=np.random.uniform(*self._env.unwrapped._arm_sampling_bounds),
                yaw=np.random.uniform(-np.pi, np.pi),
            ),
            'block_initial': self.to_pose(
                pos=info[f'privileged/block_{target_block}_pos'],
                yaw=info[f'privileged/block_{target_block}_yaw'][0],
            ),
            'block_goal': self.to_pose(
                pos=info['privileged/target_block_pos'],
                yaw=info['privileged/target_block_yaw'][0],
            ),
        }

        times, poses, grasps = self.compute_keyframes(plan_input)
        poses = [poses[name] for name in times.keys()]
        grasps = [grasps[name] for name in times.keys()]
        times = list(times.values())

        self._t_init = info['time'][0]
        self._t_max = times[-1]
        self._done = False
        self._plan = self.compute_plan(times, poses, grasps)



import os

import gymnasium
import numpy as np
from gymnasium import utils
from gymnasium.envs.mujoco import MujocoEnv
from gymnasium.spaces import Box

DEFAULT_CAMERA_CONFIG = {
    'distance': 4.0,
}


class AntEnv(MujocoEnv, utils.EzPickle):
    """Gymnasium Ant environment.

    Unlike the original Ant environment, this environment uses a restricted joint range for the actuators, as typically
    done in previous works in hierarchical reinforcement learning. It also uses a control frequency of 10Hz instead of
    20Hz, which is the default in the original environment.
    """

    xml_file = os.path.join(os.path.dirname(__file__), 'assets', 'ant.xml')
    metadata = {
        'render_modes': ['human', 'rgb_array', 'depth_array'],
        'render_fps': 10,
    }
    if gymnasium.__version__ >= '1.1.0':
        metadata['render_modes'] += ['rgbd_tuple']

    def __init__(
        self,
        xml_file=None,
        ctrl_cost_weight=0.5,
        use_contact_forces=False,
        contact_cost_weight=5e-4,
        healthy_reward=1.0,
        terminate_when_unhealthy=True,
        healthy_z_range=(0.2, 1.0),
        contact_force_range=(-1.0, 1.0),
        reset_noise_scale=0.1,
        exclude_current_positions_from_observation=True,
        **kwargs,
    ):
        if xml_file is None:
            xml_file = self.xml_file
        utils.EzPickle.__init__(
            self,
            xml_file,
            ctrl_cost_weight,
            use_contact_forces,
            contact_cost_weight,
            healthy_reward,
            terminate_when_unhealthy,
            healthy_z_range,
            contact_force_range,
            reset_noise_scale,
            exclude_current_positions_from_observation,
            **kwargs,
        )

        self._ctrl_cost_weight = ctrl_cost_weight
        self._contact_cost_weight = contact_cost_weight

        self._healthy_reward = healthy_reward
        self._terminate_when_unhealthy = terminate_when_unhealthy
        self._healthy_z_range = healthy_z_range

        self._contact_force_range = contact_force_range

        self._reset_noise_scale = reset_noise_scale

        self._use_contact_forces = use_contact_forces

        self._exclude_current_positions_from_observation = exclude_current_positions_from_observation

        obs_shape = 27
        if not exclude_current_positions_from_observation:
            obs_shape += 2
        if use_contact_forces:
            obs_shape += 84

        observation_space = Box(low=-np.inf, high=np.inf, shape=(obs_shape,), dtype=np.float64)

        MujocoEnv.__init__(
            self,
            xml_file,
            5,
            observation_space=observation_space,
            default_camera_config=DEFAULT_CAMERA_CONFIG,
            **kwargs,
        )

    @property
    def healthy_reward(self):
        return float(self.is_healthy or self._terminate_when_unhealthy) * self._healthy_reward

    def control_cost(self, action):
        control_cost = self._ctrl_cost_weight * np.sum(np.square(action))
        return control_cost

    @property
    def contact_forces(self):
        raw_contact_forces = self.data.cfrc_ext
        min_value, max_value = self._contact_force_range
        contact_forces = np.clip(raw_contact_forces, min_value, max_value)
        return contact_forces

    @property
    def contact_cost(self):
        contact_cost = self._contact_cost_weight * np.sum(np.square(self.contact_forces))
        return contact_cost

    @property
    def is_healthy(self):
        state = self.state_vector()
        min_z, max_z = self._healthy_z_range
        is_healthy = np.isfinite(state).all() and min_z <= state[2] <= max_z
        return is_healthy

    @property
    def terminated(self):
        terminated = not self.is_healthy if self._terminate_when_unhealthy else False
        return terminated

    def step(self, action):
        xy_position_before = self.get_body_com('torso')[:2].copy()
        self.do_simulation(action, self.frame_skip)
        xy_position_after = self.get_body_com('torso')[:2].copy()

        xy_velocity = (xy_position_after - xy_position_before) / self.dt
        x_velocity, y_velocity = xy_velocity

        forward_reward = x_velocity
        healthy_reward = self.healthy_reward

        rewards = forward_reward + healthy_reward

        costs = ctrl_cost = self.control_cost(action)

        terminated = self.terminated
        observation = self._get_obs()
        info = {
            'reward_forward': forward_reward,
            'reward_ctrl': -ctrl_cost,
            'reward_survive': healthy_reward,
            'x_position': xy_position_after[0],
            'y_position': xy_position_after[1],
            'distance_from_origin': np.linalg.norm(xy_position_after, ord=2),
            'x_velocity': x_velocity,
            'y_velocity': y_velocity,
            'forward_reward': forward_reward,
        }
        if self._use_contact_forces:
            contact_cost = self.contact_cost
            costs += contact_cost
            info['reward_ctrl'] = -contact_cost

        reward = rewards - costs

        if self.render_mode == 'human':
            self.render()
        return observation, reward, terminated, False, info

    def _get_obs(self):
        position = self.data.qpos.flat.copy()
        velocity = self.data.qvel.flat.copy()

        if self._exclude_current_positions_from_observation:
            position = position[2:]

        if self._use_contact_forces:
            contact_force = self.contact_forces.flat.copy()
            return np.concatenate((position, velocity, contact_force))
        else:
            return np.concatenate((position, velocity))

    def reset_model(self):
        noise_low = -self._reset_noise_scale
        noise_high = self._reset_noise_scale

        qpos = self.init_qpos + self.np_random.uniform(low=noise_low, high=noise_high, size=self.model.nq)
        qvel = self.init_qvel + self._reset_noise_scale * self.np_random.standard_normal(self.model.nv)
        self.set_state(qpos, qvel)

        observation = self._get_obs()

        return observation
